### ===========================================================================
### -package  : SimpleClass-extra
#
package require SimpleDevLibPackage 1.0
declare-package {

   SimpleClass-extra

} -version {

   1.0

} -project {

   The Simple Development Library.

} -synopsis {

   The Simple Development Library [incr Tcl] classes declaration.

} -module {

   extra

} -api {

   * K<::Simple::Class::information objects> ?-class name? ?-isa name? ?pattern?
     Returns the list of objects matching the pattern.

   * K<::Simple::Class::information aliases> class
     Returns a class list of aliases

   * K<::Simple::Class::information errors> class
     Returns a class list of errors

   * K<::Simple::Class::information exportedcommands> class
     Returns a class list of exported commands

   * K<::Simple::Class::information attributes> class ?-access choice?
     Returns a class list of attributes matching the given access level

   * K<::Simple::Class::information api> class
     Returns a class public API signature.

   * K<::Simple::Class::information commands> class
     Returns a class current list of public and exported commands.

   * K<::Simple::Class::information metadata> class
     Returns a class metadata>.

   * K<::Simple::Class::information methods> class
     Returns a class current list of public methods.

   * K<::Simple::Class::information itkoptions> class
     Returns a class list of [incr Tk] options

   * K<::Simple::Class::information itkcomponents> class
     Returns a class list of [incr Tk] components.

   * K<::Simple::Class::information classes> ?pattern?
     Returns the list of classes matching the pattern.

   * K<::Simple::Class::information exists> class
     Returns whether a class exists.

   * K<::Simple::Class::information variables> class ?-access choice?
     Returns a class list of variables matching the given access level

   * K<::Simple::ItkOption::information type> class itkoption
     Returns an [incr Tk] option type.

   * K<::Simple::ItkOption::information choices> class itkoption
     Returns an [incr Tk] option choices.

   * K<::Simple::ItkOption::information description> class itkoption
     Returns an [incr Tk] option description.

   * K<::Simple::ItkOption::information resourcename> class itkoption
     Returns an [incr Tk] option resource name.

   * K<::Simple::ItkOption::information resourceclass> class itkoption
     Returns an [incr Tk] option resource class.

   * K<::Simple::ItkOption::information defaultgiven> class itkoption
     Returns whether an [incr Tk] option has a default value.

   * K<::Simple::ItkOption::information default> class itkoption
     Returns an [incr Tk] option default value.

   * K<::Simple::ItkOption::information monitortype> class itkoption
     Returns whether an [incr Tk] option is being monitored.

   * K<::Simple::ItkComponent::information access> class itkcomponent
     Returns an [incr Tk] component access mode.

   * K<::Simple::ItkComponent::information description> class itkcomponent
     Returns an [incr Tk] component description,

   * K<::Simple::ItkComponent::information ignore> class itkcomponent
     Returns an [incr Tk] component list of ignored options.

   * K<::Simple::ItkComponent::information keep> class itkcomponent
     Returns an [incr Tk] component list of kept options.

   * K<::Simple::ItkComponent::information rename> class itkcomponent
     Returns an [incr Tk] component list of renamed options.

   * K<::Simple::ItkComponent::information resourceclass> class itkcomponent
     Returns an [incr Tk] component resource class.

} -history {

   26-oct-2004   Added A<-error> to K<declare-class>, version 0.5.5
   04-apr-2005   Added debug method and A<-debug> option, version 0.5.6
   22-jun-2005   The Simple Development Library version 1.0

} -copyright {

   Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)

} -command {

### ===========================================================================
### -command  : ::Simple::Class::information exists

{
   ::Simple::Class::information exists

} -synopsis {

   Returns whether a class exists.

} -arguments {

   { class            name           {Class name}}

} -returns {

   Whether the given class exists.

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Return whether the given class exists
   if {[catch {namespace eval $qualifiedClass info class}]} {
      set answer 0
   } else {
      set answer 1
   }
   return $answer

}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information classes

{
   ::Simple::Class::information classes

} -synopsis {

   Returns the list of classes matching the pattern.

} -overview {
 
      This procedure wraps the K<::itcl::find classes> [incr Tcl] command. The
   following is an excerpt from that command manual page:

      This procedure returns a list of classes.  Classes in the current
   namespace are listed first, followed by classes in all other namespaces in
   the interpreter.  If the optional pattern is specified, then only those
   classes whose name match the pattern are returned.
   
      If a class resides in the current namespace context, this command
   reports its unqualified.  However, if the pattern contains namespace
   qualifiers, or if the class resides in another context, this command
   reports its qualified name.  
 
} -returns {

   The list of classes matching the pattern.

} -arguments {

   {?pattern?         pattern {}     {Pattern}}

} -references {

   * The K<::itcl::find classes> [incr Tcl] command.

   * The K<itcl_info classes> [incr Tcl] command.

} -body {

   ### Return the list of classes matching the pattern
   return [uplevel 1 ::itcl::find classes $pattern]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information objects

{
   ::Simple::Class::information objects

} -synopsis {

   Returns the list of objects matching the pattern.

} -overview {
 
      This procedure wraps the K<::itcl::find classes> [incr Tcl] command. 
   The order of the arguments is different due to a limitation in The Simple
   Development Library (and Tcl) which forces the optional arguments to be the
   last ones.  The following is an excerpt from that command manual page:

      This procedure returns a list of objects.  Objects in the current
   namespace are listed first, followed by objects in all other namespaces in
   the interpreter.  If the optional pattern is specified, then only those
   objects whose name match the pattern are returned.

      If the optional A<-class> flag is given, this list is restricted to
   objects whose most-specific class is the one given as parameter to that
   flag.  If the A<-isa> flag is given, this list is further restricted to
   objects having the class given as paramter to that flag anywhere in their
   heritage.
   
      If an object resides in the current namespace context, this command
   reports its unqualified name.  However, if the pattern contains namespace
   qualifiers, or if the object resides in another context, this command
   reports its fully-qualified name.
 
} -returns {

   The list of objects matching the pattern.

} -arguments {

   {-class            name %DEFAULT% {Class name}}
   {-isa              name %DEFAULT% {Base class name}}
   {?pattern?         pattern {}     {Pattern}}

} -references {

   * The K<::itcl::find objects> [incr Tcl] command.

   * The K<itcl_info objects> [incr Tcl] command.

} -body {

   # Compose the command
   set command "::itcl::find objects $pattern"
   if {[::Simple::Argument::information flaggiven -class]} {
      append command " -class $class"
   }
   if {[::Simple::Argument::information flaggiven -isa]} {
      append command " -isa $isa"
   }

   ### Return the list of objects matching the pattern
   return [uplevel 1 $command]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information metadata

{
   ::Simple::Class::information metadata

} -synopsis {

   Returns a class metadata.

} -overview {
 
      This procedure returns a list of pairs suitable to be given as argument
   to the K<array set> command.  Each element is one of the flags given to
   K<declare-class> upon class declaration (with no hypen), but for the
   A<-api> and A<-body> flags (which are not stored), and the item
   declarations (A<-error>, A<-variable>, A<-attribute>, A<-itkoption>,
   A<-itkcomponent>, A<-command> and A<-method>); these are obtained using the
   corresponding K<::Simple::Class::information> subcommand.

      Thus, the actual list of flags is:  A<-version>, A<-project>,
   A<-synopsis>, A<-overview>, A<-usage>, A<-keywords>, A<-examples>,
   A<-classtype>, A<-stereotype>, A<-extends>, A<-assumptions>, A<-resources>,
   A<-limitations>, A<-effects>, A<-details>, A<-references>, A<-remarks>,
   A<-todo>, A<-history> and A<-copyright>.
 
      Notice that no data is returned unless The Simple Development Library
   A<-storemetadata> option is set when the class is declared. 
   
      If a particular flag was not given in the class declaration, the empty
   string is returned as that flag contents.
 
} -returns {

   The class metadata as an item/contents pairs list.

} -arguments {

   { class            name           {Class name}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### The class metadata was stored
   set answer [list]
   set metadata [set ${qualifiedClass}::%Metadata%]
   if {[llength $metadata] > 0} {
      foreach flag [list version project synopsis overview usage\
         keywords examples classtype stereotype extends assumptions\
         resources limitations effects details references remarks todo\
         history copyright] contents $metadata {

         ### Add the metadata to the list
         lappend answer $flag $contents
      }
   }

   ### Return the class metadata
   return $answer
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::assert-class-declared

{
   ::Simple::Class::Priv::assert-class-declared

} -synopsis {

   Throws an error if a class has not been declared.

} -arguments {

   { class            qualifiedname  {Class name}}

} -body {

   if {![info exists ${class}::%Metadata%]} {
      ::Simple::Error::throw ::Simple::Class::UNDECLARED $class
   }
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information errors

{
   ::Simple::Class::information errors

} -synopsis {

   Returns a class list of errors

} -returns {

   The class list of errors

} -arguments {

   { class            name           {Class name}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Return the class list of errors
   set answer [list]
   foreach error [array names ${qualifiedClass}::%Errors%] {
      lappend answer ${qualifiedClass}::$error
   }
   return $answer
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information exportedcommands

{
   ::Simple::Class::information exportedcommands

} -synopsis {

   Returns a class list of exported commands

} -returns {

   The class list of exported commands

} -arguments {

   { class            name           {Class name}}

} -references {
 
   * The K<::Simple::Package::information exportedcommands> command.

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Return the class list of exported commands
   return [set ${qualifiedClass}::%ExportedCommands%]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information aliases

{
   ::Simple::Class::information aliases

} -synopsis {

   Returns a class list of aliases

} -returns {

   The class list of aliases

} -arguments {

   { class            name           {Class name}}

} -references {
 
   * The K<::Simple::Package::information aliases> command.

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Return the class list of aliases
   return [set ${qualifiedClass}::%Aliases%]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information variables

{
   ::Simple::Class::information variables

} -synopsis {

   Returns a class list of variables matching the given access level

} -returns {

   The class list of variables matching the given access level

} -arguments {

   { class            name           {Class name}}
   {-access           choice {public protected private}
                                     {Variables access mode}}

} -body {
   return [::Simple::Class::Priv::information-variables-or-attributes\
      $class $access common]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information attributes

{
   ::Simple::Class::information attributes

} -synopsis {

   Returns a class list of attributes matching the given access level

} -returns {

   The class list of attributes matching the given access level

} -arguments {

   { class            name           {Class name}}
   {-access           choice {public protected private}
                                     {Attributes access mode}}

} -body {
   return [::Simple::Class::Priv::information-variables-or-attributes\
      $class $access variable]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::information-variables-or-attributes

{
   ::Simple::Class::Priv::information-variables-or-attributes

} -synopsis {

   Returns a class current list of variables or attributes matching the given access level.

} -returns {

   The class current list of variables or attributes matching the given access level.

} -arguments {

   { class            name           {Class name}}
   { access           choice {public protected private}
                                     {Variable access mode}}
   { commonOrVariable choice {common variable}
                                     {Whether to return variables (common) or
                                      attributes (variable)}}
} -remarks {

   * The returned list does not contain items with a percent sign in its name
     are those are assumed to be internal to The Simple Development Library.

   * The returned list does not contain the V<this> variable.

   * For derived classes, the returned list contains only those private
     elements of the actual class and not those of any possible base ones.
     On the contrary, [incr Tcl] returns all private items, even those
     belonging to base classes.

} -todo {

   * All the class elements are static, and so the "current list" concept
     does not really apply here:  it is possible to store a class list of
     variables and attributes instead of dynamically retrieving it.

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class 4] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Loop over variables and attributes
   set answer [list]
   set isNotPrivate [string compare $access private]
   foreach item [namespace eval $qualifiedClass info variable] {

      ### The variable or attribute is not internal, and is not the V<this>
      ###    variable, and  for private items, the item belongs to the class
      ###    and not to any possible base one
      if {[::Simple::Priv::is-not-internal $item] &&\
         ($isNotPrivate || [string match ${qualifiedClass}::* $item]) &&\
         ![string match *::this $item]} {

         ### The variable or attribute matches the given access level
         foreach {itemType itemAccess} [namespace eval $qualifiedClass\
            info variable $item -type -protection] break
         if {![string compare $itemType $commonOrVariable] &&\
            ![string compare $itemAccess $access]} {

            ### Add the variable or attribute
            lappend answer $item
         }
      }
   }

   ### Return the class current list of variables or attributes matching the given access level
   return $answer
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information itkoptions

{
   ::Simple::Class::information itkoptions

} -synopsis {

   Returns a class list of [incr Tk] options

} -returns {

   The class list of [incr Tk] options

} -arguments {

   { class            name           {Class name}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Return the class list of [incr Tk] options
   return [array names ${qualifiedClass}::%ItkOptions%]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkOption::information type

{
   ::Simple::ItkOption::information type

} -synopsis {

   Returns an [incr Tk] option type.

} -returns {

   The [incr Tk] option type.

} -arguments {

   { class            name           {Class name}}
   { itkoption        flag           {[incr Tk] option}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] option exists
   if {![info exists ${qualifiedClass}::%ItkOptions%($itkoption)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {an option} $itkoption
   }

   ### Return the [incr Tk] option type
   return [lindex [set ${qualifiedClass}::%ItkOptions%($itkoption)] 0]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkOption::information choices

{
   ::Simple::ItkOption::information choices

} -synopsis {

   Returns an [incr Tk] option choices.

} -returns {

   The [incr Tk] option choices.

} -arguments {

   { class            name           {Class name}}
   { itkoption        flag           {[incr Tk] option}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] option exists
   if {![info exists ${qualifiedClass}::%ItkOptions%($itkoption)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {an option} $itkoption
   }

   ### Return the [incr Tk] option choices
   return [lindex [set ${qualifiedClass}::%ItkOptions%($itkoption)] 1]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkOption::information description

{
   ::Simple::ItkOption::information description

} -synopsis {

   Returns an [incr Tk] option description.

} -returns {

   The [incr Tk] option description.

} -arguments {

   { class            name           {Class name}}
   { itkoption        flag           {[incr Tk] option}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] option exists
   if {![info exists ${qualifiedClass}::%ItkOptions%($itkoption)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {an option} $itkoption
   }

   ### Return the [incr Tk] option description
   return [lindex [set ${qualifiedClass}::%ItkOptions%($itkoption)] 2]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkOption::information resourcename

{
   ::Simple::ItkOption::information resourcename

} -synopsis {

   Returns an [incr Tk] option resource name.

} -returns {

   The [incr Tk] option resource name.

} -arguments {

   { class            name           {Class name}}
   { itkoption        flag           {[incr Tk] option}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] option exists
   if {![info exists ${qualifiedClass}::%ItkOptions%($itkoption)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {an option} $itkoption
   }

   ### Return the [incr Tk] option resource name
   return [lindex [set ${qualifiedClass}::%ItkOptions%($itkoption)] 3]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkOption::information resourceclass

{
   ::Simple::ItkOption::information resourceclass

} -synopsis {

   Returns an [incr Tk] option resource class.

} -returns {

   The [incr Tk] option resource class.

} -arguments {

   { class            name           {Class name}}
   { itkoption        flag           {[incr Tk] option}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] option exists
   if {![info exists ${qualifiedClass}::%ItkOptions%($itkoption)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {an option} $itkoption
   }

   ### Return the [incr Tk] option resource class
   return [lindex [set ${qualifiedClass}::%ItkOptions%($itkoption)] 4]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkOption::information defaultgiven

{
   ::Simple::ItkOption::information defaultgiven

} -synopsis {

   Returns whether an [incr Tk] option has a default value.

} -arguments {

   { class            name           {Class name}}
   { itkoption        flag           {[incr Tk] option}}

} -returns {

   Whether the [incr Tk] option has a default value.

} -references {
 
   * The K<::Simple::Variable::information defaultgiven> procedure.
 
   * The K<::Simple::Option::information defaultgiven> procedure.

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] option exists
   if {![info exists ${qualifiedClass}::%ItkOptions%($itkoption)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {an option} $itkoption
   }

   ### Return whether the [incr Tk] option has a default value
   return [lindex [set ${qualifiedClass}::%ItkOptions%($itkoption)] 5]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkOption::information default

{
   ::Simple::ItkOption::information default

} -synopsis {

   Returns an [incr Tk] option default value.

} -arguments {

   { class            name           {Class name}}
   { itkoption        flag           {[incr Tk] option}}

} -returns {

   The [incr Tk] option default value.

} -references {
 
   * The K<::Simple::Variable::information default> procedure.
 
   * The K<::Simple::Option::information default> procedure.

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] option exists
   if {![info exists ${qualifiedClass}::%ItkOptions%($itkoption)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {an option} $itkoption
   }

   ### Assert the [incr Tk] option has a default value
   if {![lindex [set ${qualifiedClass}::%ItkOptions%($itkoption)] 5]} {
      ::Simple::Error::throw ::Simple::Variable::NO-DEFAULT-VALUE\
         {[incr Tk] option} $itkoption
   }

   ### Return the [incr Tk] option default value
   return [lindex [set ${qualifiedClass}::%ItkOptions%($itkoption)] 6]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkOption::information monitortype

{
   ::Simple::ItkOption::information monitortype

} -synopsis {

   Returns whether an [incr Tk] option is being monitored.

} -arguments {

   { class            name           {Class name}}
   { itkoption        flag           {[incr Tk] option}}

} -returns {

   Whether an [incr Tk] option is being monitored.

} -references {
 
   * The K<::Simple::Variable::information monitortype> procedure.
 
} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] option exists
   if {![info exists ${qualifiedClass}::%ItkOptions%($itkoption)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {an option} $itkoption
   }

   ### Return whether the [incr Tk] option is being monitored
   #   to ensure its values conform to its type
   return [lindex [set ${qualifiedClass}::%ItkOptions%($itkoption)] 7]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information itkcomponents

{
   ::Simple::Class::information itkcomponents

} -synopsis {

   Returns a class list of [incr Tk] components.

} -returns {

   The class list of [incr Tk] components

} -arguments {

   { class            name           {Class name}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Return the class list of [incr Tk] components
   return [array names ${qualifiedClass}::%ItkComponents%]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkComponent::information access

{
   ::Simple::ItkComponent::information access

} -synopsis {

   Returns an [incr Tk] component access mode.

} -returns {

   The [incr Tk] component access mode.

} -arguments {

   { class            name           {Class name}}
   { itkcomponent     unqualifiedname
                                     {[incr Tk] component}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] component exists
   if {![info exists ${qualifiedClass}::%ItkComponents%($itkcomponent)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {a component} $itkcomponent
   }

   ### Return the [incr Tk] component access mode
   return [lindex [set ${qualifiedClass}::%ItkComponents%($itkcomponent)] 0]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkComponent::information description

{
   ::Simple::ItkComponent::information description

} -synopsis {

   Returns an [incr Tk] component description.

} -returns {

   The [incr Tk] component description.

} -arguments {

   { class            name           {Class name}}
   { itkcomponent     unqualifiedname
                                     {[incr Tk] component}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] component exists
   if {![info exists ${qualifiedClass}::%ItkComponents%($itkcomponent)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {a component} $itkcomponent
   }

   ### Return the [incr Tk] component description
   return [lindex [set ${qualifiedClass}::%ItkComponents%($itkcomponent)] 1]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkComponent::information ignore

{
   ::Simple::ItkComponent::information ignore

} -synopsis {

   Returns an [incr Tk] component list of ignored options.

} -returns {

   The [incr Tk] component list of ignored options.

} -arguments {

   { class            name           {Class name}}
   { itkcomponent     unqualifiedname
                                     {[incr Tk] component}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] component exists
   if {![info exists ${qualifiedClass}::%ItkComponents%($itkcomponent)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {a component} $itkcomponent
   }

   ### Return the [incr Tk] component list of ignored options
   return [lindex [set ${qualifiedClass}::%ItkComponents%($itkcomponent)] 2]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkComponent::information keep

{
   ::Simple::ItkComponent::information keep

} -synopsis {

   Returns an [incr Tk] component list of kept options.

} -returns {

   The [incr Tk] component list of kept options.

} -arguments {

   { class            name           {Class name}}
   { itkcomponent     unqualifiedname
                                     {[incr Tk] component}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] component exists
   if {![info exists ${qualifiedClass}::%ItkComponents%($itkcomponent)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {a component} $itkcomponent
   }

   ### Return the [incr Tk] component list of kept options
   return [lindex [set ${qualifiedClass}::%ItkComponents%($itkcomponent)] 3]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkComponent::information rename

{
   ::Simple::ItkComponent::information rename

} -synopsis {

   Returns an [incr Tk] component list of renamed options.

} -returns {

   The [incr Tk] component list of renamed options.

} -arguments {

   { class            name           {Class name}}
   { itkcomponent     unqualifiedname
                                     {[incr Tk] component}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] component exists
   if {![info exists ${qualifiedClass}::%ItkComponents%($itkcomponent)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {a component} $itkcomponent
   }

   ### Return the [incr Tk] component list of renamed options
   return [lindex [set ${qualifiedClass}::%ItkComponents%($itkcomponent)] 4]
}
} -command {

### ===========================================================================
### -command  : ::Simple::ItkComponent::information resourceclass

{
   ::Simple::ItkComponent::information resourceclass

} -synopsis {

   Returns an [incr Tk] component resource class.

} -returns {

   The [incr Tk] component resource class.

} -arguments {

   { class            name           {Class name}}
   { itkcomponent     unqualifiedname
                                     {[incr Tk] component}}

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Assert the [incr Tk] component exists
   if {![info exists ${qualifiedClass}::%ItkComponents%($itkcomponent)]} {
      ::Simple::Error::throw ::Simple::Class::NON-EXISTING-ITKITEM\
         $class {a component} $itkcomponent
   }

   ### Return the [incr Tk] component resource class
   return [lindex [set ${qualifiedClass}::%ItkComponents%($itkcomponent)] 5]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information commands

{
   ::Simple::Class::information commands

} -synopsis {

   Returns a class current list of public and exported commands.

} -returns {

   The class current list of public and exported commands

} -arguments {

   { class            name           {Class name}}

} -references {
 
   * The K<::Simple::Package::information api> command.
 
   * The K<::Simple::Package::information commands> command.

   * The K<::Simple::Class::information methods> command.
 
} -remarks {
 
   * The returned list does not contain base commands.
 
   * The returned list does not contain commands with a percent sign in its
     name are those are assumed to be internal to The Simple Development
     Library.
 
} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Loop over class commands and methods
   set answer [list]
   foreach commandOrMethod [namespace eval $qualifiedClass\
      {[uplevel 1 set qualifiedClass]::info function}] {

      # This class command or method and not internal
      if {[string match $qualifiedClass* $commandOrMethod] &&
         [::Simple::Priv::is-not-internal $commandOrMethod]} {

         # Class command
         foreach {type access body} [namespace eval $qualifiedClass\
            {[uplevel 1 set class]::info function\
            [uplevel 1 set commandOrMethod] -type -protection -body}] break
         if {![string compare $type proc]} {

            # The command is public or exported and is not a base command
            if {(![string compare $access public] ||\
               ![string compare $access exported]) &&
               ([string first $::Simple::BaseCommandTag $body] == -1)} {

               lappend answer $commandOrMethod
            }
         } 
      }
   }

   ### Return the class list of public and exported commands
   return $answer
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information methods

{
   ::Simple::Class::information methods

} -synopsis {

   Returns a class current list of public methods.

} -returns {

   The class current list of public methods

} -arguments {

   { class            name           {Class name}}

} -references {
 
   * The K<::Simple::Package::information api> command.
 
   * The K<::Simple::Package::information commands> command.

} -remarks {
 
   * The returned list does not contain base methods.
 
   * The returned list does not contain methods with a percent sign in its
     name are those are assumed to be internal to The Simple Development
     Library.
 
} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Loop over class commands and methods
   set answer [list]
   foreach commandOrMethod [namespace eval $qualifiedClass\
      {[uplevel 1 set qualifiedClass]::info function}] {

      # This class command or method and not internal
      if {[string match $qualifiedClass* $commandOrMethod] &&
         [::Simple::Priv::is-not-internal $commandOrMethod]} {

         # Class method
         foreach {type access body} [namespace eval $qualifiedClass\
            {[uplevel 1 set class]::info function\
            [uplevel 1 set commandOrMethod] -type -protection -body}] break
         if {![string compare $type method]} {

            # The method is public and is not an [incr Tcl]
            # builtin nor a base method
            if {![string compare $access public] &&
               ![string match @itcl-builtin-* $body] &&
               ([string first $::Simple::BaseCommandTag $body] == -1)} {

               lappend answer $commandOrMethod
            }
         } 
      }
   }

   ### Return the class list of public methods
   return $answer
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::information api

{
   ::Simple::Class::information api

} -synopsis {

   Returns a class public API signature.

} -overview {

      This procedure returns a string containing a class public API
   signatures.
 
      The class public and exported commands are included, as well as the
   public constructor, public destructor and public methods.
 
} -returns {

   The given class public API signatures as follows:

#verbatim
   (extends <class list>)
   Class constructor:
      <class 1 constructor API>
   Class destructor:
      <class 1 destructor API>
   Class commands:
      <class 1 command 1 API>
      <class 1 command 2 API>
      ...
   Class methods:
      <class 1 method 1 API>
      <class 1 method 2 API>
      ...
#end

} -arguments {

   { class            name           {Class name}}

} -references {
 
   * The K<::Simple::Package::information api> command.
 
   * The K<::Simple::Class::information commands> command.

   * The K<::Simple::Class::information methods> command.

} -remarks {
 
   * The returned list does not contain base methods.
 
   * The returned list does not contain methods with a percent sign in its
     name are those are assumed to be internal to The Simple Development
     Library.
 
} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Assert the class has been declared
   ::Simple::Class::Priv::assert-class-declared $qualifiedClass

   ### Add the extended classes list
   set answer {}
   set extends [namespace eval $qualifiedClass {[uplevel 1 set class]::info inherit}]
   if {[llength $extends] > 0} {
      append answer\
         "(extends [::Simple::List::enumerate $extends and])\n"
   }

   ### Loop over the class public or exported commands
   set classCommandsAPIs [list]
   set constructorAPI {}
   set destructorAPI {}
   foreach classCommand\
      [lsort [::Simple::Class::information commands $qualifiedClass]] {

      ### Constructor class command
      set classCommandAPI [::Simple::Declare::Priv::command-api\
         $classCommand [set ${qualifiedClass}::%ExportedCommands%]\
         [set ${qualifiedClass}::%Aliases%]]
      if {![string compare $classCommand ${class}::constructor-command]} {

         ### Set the constructor API
         set constructorAPI $classCommandAPI

      ### Destructor class command
      } elseif {![string compare $classCommand ${class}::destructor-command]} {

         ### Set the destructor API
         set destructorAPI $classCommandAPI

      ### Other class commands
      } else {

         ### Add the class command API
         lappend classCommandsAPIs $classCommandAPI
      }
   }

   ### Loop over the class public methods
   set methodsAPIs [list]
   foreach classMethod\
      [lsort [::Simple::Class::information methods $qualifiedClass]] {

      ### Constructor
      set unqualifiedMethodAPI [string trimright "[namespace tail\
         $classMethod] [{::Simple::Proc::information usage} $classMethod]"]
      if {[string match constructor* $unqualifiedMethodAPI]} {

         ### Set the constructor API
         set constructorAPI\
            "$class <object> [string range $unqualifiedMethodAPI 12 end]"

      ### Destructor
      } elseif {[string match destructor* $unqualifiedMethodAPI]} {

         ### Set the destructor API
         set destructorAPI\
            "<object> [string range $unqualifiedMethodAPI 11 end]"

      ### Other methods
      } else {

         ### Add the method API
         lappend methodsAPIs "<object> $unqualifiedMethodAPI"
      }
   }

   ### Add the constructor, destructor, class commands and methods APIs
   if {[string compare $constructorAPI {}]} {
      append answer "Class constructor:\n   $constructorAPI\n"
   }
   if {[string compare $destructorAPI {}]} {
      append answer "Class destructor:\n   $destructorAPI\n"
   }
   if {[llength $classCommandsAPIs] > 0} {
      append answer "Class commands:\n"
      foreach classCommandAPI [lsort $classCommandsAPIs] {
         append answer "   $classCommandAPI\n"
      }
   }
   if {[llength $methodsAPIs] > 0} {
      append answer "Class methods:\n"
      foreach methodAPI [lsort $methodsAPIs] {
         append answer "   $methodAPI\n"
      }
   }

   ### Return the given class public API signatures
   return [string trimright $answer \n]
}
}
