### ===========================================================================
### -package  : SimpleFile
#
package require SimpleDevLibPackage 1.0
declare-package {

   SimpleFile

} -version {

   1.0

} -project {

   The Simple Development Library.

} -synopsis {

   The Simple Development Library file handling.

} -module {

   implementation

} -paradigm {

   procedural

} -overview {

    This package provides procedures for reading and writing files.

} -usage {

       Use P<::Simple::File::swallow> to read data from a file and
    P<::Simple::File::dump> to write data into a file.

} -keywords {

   file

} -examples {

   # Install the package
   package require SimpleDevLibPackage 1.0
   ::Simple::Package::require-and-install SimpleFile

   # Read a file into a variable
   set fileContents [::Simple::File::swallow file.dat]

   # Modify the data
   append fileContents "\nNew line"

   # Write the modified data back
   ::Simple::File::dump file.dat $fileContents

} -requisites {

   {SimpleDeclare 1.0}

} -namespaces {

   ::Simple::File

} -api {

   * K<::Simple::File::swallow> fileName
     Reads a file.
 
   * K<::Simple::File::dump> fileName fileContents 
     Writes a file.

} -history {

   17-feb-2003   First public release, version 0.5
   22-jun-2005   The Simple Development Library version 1.0

} -copyright {

   Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)

} -error {

   {
      ::Simple::File::NON-READABLE-FILE
   } -message {
      file "%s" is not readable
   } -explanation {
      The file <file> is not readable for this user
   } -corrective {
      Change the file permissions
   }

} -error {

   {
      ::Simple::File::CANT-OPEN-FILE
   } -message {
      could not open file "%s" for %s
   } -explanation {
      The file <file> could not be opened for <reading or wirintg> by this user
   } -corrective {
      Change the file and/or directory permissions
   }

} -error {

   {
      ::Simple::File::CANT-READ-FILE
   } -message {
      could not read file "%s"
   } -explanation {
      There was an error reading the file <file>
   } -corrective {
      Check the file
   }

} -error {

   {
      ::Simple::File::CANT-WRITE-FILE
   } -message {
      could not write file "%s"
   } -explanation {
      There was an error writing the file <file>
   } -corrective {
      Check the file and directory
   }

} -error {

   {
      ::Simple::File::CANT-CLOSE-FILE
   } -message {
      could not close file "%s"
   } -explanation {
      There was an error clsoing the file <file>
   } -corrective {
      Ensure there is enough free space on the device
   }

} -command {

### ===========================================================================
### -command  : ::Simple::File::swallow

{
   ::Simple::File::swallow

} -synopsis {

   Reads a file.

} -overview {

      This procedure reads a file and returns its contents.  The file must
   exist and must be readable.

} -arguments {
   { fileName           string               {File name}} 
} -returns {

   The file contents.

} -body {

   ### File does not exist
   if {![file exists $fileName]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::NON-EXISTING file $fileName 

   ### File exists
   } else {

      ### File is not readable
      if {![file readable $fileName]} {

         ### Throw error
         ::Simple::Error::throw ::Simple::File::NON-READABLE-FILE $fileName 

      ### File is readable
      } else {
      
         ### Cannot open the file
         if {[catch {set fileId [open $fileName r]}]} {

            ### Throw error
            ::Simple::Error::throw\
               ::Simple::File::CANT-OPEN-FILE $fileName reading

         ### Can open the file
         } else {
              
            ### Cannot read the file
            if {[catch {
               set fileContents [read $fileId [file size $fileName]]
            }]} {

               ### Throw error
               ::Simple::Error::throw\
                  ::Simple::File::CANT-READ-FILE $fileName 
               
            ### Can read the file
            } else {

               ### Cannot close the file
               if {[catch {close $fileId}]} {

                  ### Throw error
                  ::Simple::Error::throw\
                     ::Simple::File::CANT-CLOSE-FILE $fileName 
               }
            }
         }
      }
   }

   ### Return the file contents
   return $fileContents
}

} -command {

### ===========================================================================
### -command  : ::Simple::File::dump

{
   ::Simple::File::dump

} -synopsis {

   Writes a file.

} -overview {

      This procedure writes data into a file.  If the file exists it is
   truncated. The file must be writable.

} -arguments {
   { fileName           string               {File name}} 
   { fileContents       string               {Data to be written into
                                              the file}}
} -returns {

   The empty string.

} -body {

   ### Cannot open the file
   if {[catch {set fileId [open $fileName w]}]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::File::CANT-OPEN-FILE $fileName writing

   ### Can open the file
   } else {

      ### Cannot write into the file
      if {[catch {puts -nonewline $fileId $fileContents}]} {

         ### Throw error
         ::Simple::Error::throw ::Simple::File::CANT-WRITE-FILE $fileName 
            
      ### Can write into the file
      } else {

         ### Cannot close the file
         if {[catch {close $fileId}]} {

            ### Throw error
            ::Simple::Error::throw ::Simple::File::CANT-CLOSE-FILE $fileName 
         }
      }
   }

   ### Return the empty string
   return {}
}
}
