### ===========================================================================
### -package  : SimpleOption-extra
#
package require SimpleDevLibPackage 1.0
declare-package {

   SimpleOption-extra

} -version {

   1.0

} -project {

   The Simple Development Library.

} -synopsis {

   The Simple Development Library command line options parser.

} -module {

   extra

} -api {

   * K<::Simple::Option::reset>
     Resets the declared options and usage information.

   * K<::Simple::Option::information> declared
     Returns the list of declared options.

} -history {

   11-jul-2003   Package renamed to <SimpleOption>, version 0.5.2
   22-jun-2005   The Simple Development Library version 1.0

} -copyright {

   Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)

} -command {

### ===========================================================================
### -command  : ::Simple::Option::Priv::declare-usage-group-2

{
   ::Simple::Option::Priv::declare-usage-group-2

} -synopsis {

   Specifies major groups usage.

} -arguments {
   { majorGroups      list           {Major option groups}}
   { requiredOptions  list           {Required options list}}
   { groupSpecifiedOptions
                      list           {Usage group specified options list}}
   { exclusiveGroup   boolean        {Whether the major groups are exclusive}}
} -returns {

   A list with the following elements:

#table;delimiter=":";ignore="true"
    0: the group required options.
    1: the group specified options.
#end

} -body {

   ### Loop over major groups
   set exclusiveGroupRequiredOptions [list]
   set majorGroupsSpecifiedOptions [list]
   set exclusiveGroupHasRequiredOptions 1
   foreach majorGroup $majorGroups {

      ### Assert no exclusive major group is an exclusive group
      if {$exclusiveGroup && ![string compare [lindex $majorGroup 0] X]} {
         ::Simple::Error::throw\
            ::Simple::Option::EXCLUSIVE-GROUP-AS-ALTERNATIVE $majorGroup
      }

      ### Specify the major group usage
      foreach {majorGroupRequiredOptions majorGroupSpecifiedOptions}\
         [::Simple::Option::Priv::declare-usage-group-1 $majorGroup\
         $requiredOptions $exclusiveGroup] break

      ### Exclusive group
      if {$exclusiveGroup} {

         ### Get the major group required options
         if {$exclusiveGroupHasRequiredOptions} {
            if {[llength $majorGroupRequiredOptions] == 0} {
               set exclusiveGroupHasRequiredOptions 0
            } else {
               lappend exclusiveGroupRequiredOptions\
                  [lindex $majorGroupRequiredOptions 0]
            }
         }

         ### Add the major group specified options to the
         ###    exclusive groups sepecified options list
         lappend majorGroupsSpecifiedOptions $majorGroupSpecifiedOptions

         ### Add the alternation symbol to the usage string
         append ::Simple::Option::Usage {| }

      ### Non-exclusive group
      } else {

         ### Add the major group required options to the required options list
         set requiredOptions\
            [concat $requiredOptions $majorGroupRequiredOptions]

         ### Add the major group required options to each
         ###    option in the group required options list
         foreach specifiedOption $groupSpecifiedOptions {
            set ::Simple::Option::RequiredOptions($specifiedOption)\
               [concat $::Simple::Option::RequiredOptions($specifiedOption)\
               $majorGroupRequiredOptions]
         }
      }

      ### Add the major group specified options
      ###    to the group specified options list
      set groupSpecifiedOptions\
         [concat $groupSpecifiedOptions $majorGroupSpecifiedOptions]
   }

   ### Exclusive group
   if {$exclusiveGroup} {

      ### Add other major groups specified options to
      ###    each option excluded options list
      set firstIndex 0
      set lastIndex -1
      upvar ::Simple::Option::ExcludedOptions ExcludedOptions
      foreach majorGroupSpecifiedOptions $majorGroupsSpecifiedOptions {
         set nGroupSpecifiedOptions [llength $majorGroupSpecifiedOptions]
         incr lastIndex $nGroupSpecifiedOptions
         set majorGroupExcludedOptions\
            [lreplace $groupSpecifiedOptions $firstIndex $lastIndex]
         foreach majorGroupSpecifiedOption $majorGroupSpecifiedOptions {
            if {[info exists ExcludedOptions($majorGroupSpecifiedOption)]} {
               set ExcludedOptions($majorGroupSpecifiedOption)\
                  [concat $ExcludedOptions($majorGroupSpecifiedOption)\
                  $majorGroupExcludedOptions]
            } else {
               set ExcludedOptions($majorGroupSpecifiedOption)\
                  $majorGroupExcludedOptions
            }
         }
         incr firstIndex $nGroupSpecifiedOptions
      }

      ### The new required options list is all major groups first options
      if {$exclusiveGroupHasRequiredOptions} {
         set requiredOptions [list $exclusiveGroupRequiredOptions]
      } else {
         set requiredOptions [list]
      }
   }

   ### Return the required options and the group specified options
   return [list $requiredOptions $groupSpecifiedOptions]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::reset

{
   ::Simple::Option::reset

} -synopsis {

   Resets the declared options and usage information.

} -returns {

   The empty string.

} -body {

   ::Simple::Option::Priv::reset-declare
   ::Simple::Option::Priv::reset-declare-usage

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::Priv::reset-declare

{
   ::Simple::Option::Priv::reset-declare

} -synopsis {

   Resets previous calls to K<::Simple::Option::declare>.

} -returns {

   The empty string.

} -body {

   set ::Simple::Option::PrimaryOptions [list]
   catch {unset ::Simple::Option::Options}
   catch {unset ::Simple::Option::DefaultValues}
   catch {unset ::Simple::Option::ActualValues}

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::Priv::extract-double-hyphen-values

{
   ::Simple::Option::Priv::extract-double-hyphen-values

} -synopsis {

   Extracts the "--" option values

} -arguments {
   { doubleHyphenIndex   
                      integer        {Index of "--" in the values list}}
} -returns {

   The empty string.

} -body {

   ### Link to variables in the calling scope
   upvar commandLine commandLine
   upvar values values

   ### Assert "--" is a primary declared option flag
   if {![info exists ::Simple::Option::Options(--)]} {
      if {$doubleHyphenIndex == -1} {
         ::Simple::Error::throw ::Simple::Option::EXTRA-VALUES $commandLine
      } else {
         ::Simple::Error::throw ::Simple::Option::DOUBLE-HYPHEN-NOT-ALLOWED
      }
   }

   ### Extract the "--" values
   #   Up to the last remaining argument
   set values [lrange $commandLine [incr doubleHyphenIndex 1] end]
   set commandLine [lrange $commandLine 0 [incr doubleHyphenIndex -2]]

   ### Check the "--" values
   ::Simple::Option::Priv::check-values -- $values

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information declared

{
   ::Simple::Option::information declared

} -synopsis {

   Returns the list of declared options.

} -overview {

   This procedure returns the list of primary declared option flags.

} -returns {

   The list of primary declared option flags.

} -body {

   ### Return the list of primary declared option flags
   return $::Simple::Option::PrimaryOptions

}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information secondary

{
   ::Simple::Option::information secondary

} -synopsis {

   Returns an option secondary flag.

} -overview {

      This procedure returns an option secondary flag.  The option flag
   must be primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   The option secondary flag.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Return the option secondary flag
   return [lindex $::Simple::Option::Options($option) 1]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information name

{
   ::Simple::Option::information name

} -synopsis {

   Returns an option name.

} -overview {

   This procedure returns an option name.  The option flag must be primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   The option name.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Get the option name
   set answer [lindex $::Simple::Option::Options($option) 2]
   if {![string compare $answer %DEFAULT%]} {
      set answer {}
   }

   ### Return the option name
   return $answer
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information type

{
   ::Simple::Option::information type

} -synopsis {

   Returns an option type.

} -overview {

   This procedure returns an option type.  The option flag must be primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   The option type.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Return the option type
   return [lindex $::Simple::Option::Options($option) 3]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information choices

{
   ::Simple::Option::information choices

} -synopsis {

   Returns an option choices.

} -overview {

      This procedure returns an option choices.  The option flag must be
   primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   The option choices.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Return the option choices
   return [lindex $::Simple::Option::Options($option) 4]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information range

{
   ::Simple::Option::information range

} -synopsis {

   Returns an option range.

} -overview {

   This procedure returns an option range.  The option flag must be primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   The option range.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Return the option range
   return [lindex $::Simple::Option::Options($option) 5]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information values

{
   ::Simple::Option::information values

} -synopsis {

   Returns an option number of values.

} -overview {

      This procedure returns an option number of values.  The option flag must
   be primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   The option values.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Return the option number of values
   return [lindex $::Simple::Option::Options($option) 6]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information description

{
   ::Simple::Option::information description

} -synopsis {

   Returns an option description.

} -overview {

      This procedure returns an option description.  The option flag must be
   primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   The option description.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Return the option description
   return [lindex $::Simple::Option::Options($option) 7]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information default

{
   ::Simple::Option::information default

} -synopsis {

   Returns an option default value.

} -overview {

      This procedure returns an option default value.  The option flag must be
   primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   The option default value.

} -references {
 
   * The K<::Simple::Variable::information default> procedure.
 
   * The K<::Simple::ItkOption::information default> procedure.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### The option has a default value
   set answer {}
   if {[info exists ::Simple::Option::DefaultValues($option)]} {

      ### Get it
      set answer $::Simple::Option::DefaultValues($option)

   ### The option has no default value
   } else {

      ### Throw error
      ::Simple::Error::throw ::Simple::Variable::NO-DEFAULT-VALUE\
         option $option
   }

   ### Return the option default value
   return $answer
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information defaultgiven

{
   ::Simple::Option::information defaultgiven

} -synopsis {

   Returns whether an option has a default value.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   Whether the option has default value.

} -references {
 
   * The K<::Simple::Variable::information defaultgiven> procedure.
 
   * The K<::Simple::ItkOption::information defaultgiven> procedure.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Return whether the option has a default value
   return [info exists ::Simple::Option::DefaultValues($option)]
}
}
