### ===========================================================================
### -package  : SimplePackage-extra
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library packages handling.
# -module     : extra
# -api        :
#
#  * K<::Simple::Package::information known> ?pattern?
#    Returns the list of known packages.
#
#  * K<::Simple::Package::information declared> ?pattern?
#    Returns the list of declared packages.
#
#  * K<::Simple::Package::information uninstalled> ?pattern?
#    Returns the list of uninstalled packages.
#
#  * K<::Simple::Package::information installed> ?pattern?
#    Returns the list of installed packages.
#
#  * K<::Simple::Package::information namespaces> package ?version?
#    Returns a declared package namespaces.
#
#  * K<::Simple::Package::information packagealias> alias
#    Returns the declared package an alias belongs to.
#
#  * K<::Simple::Package::information packagecommand> command
#    Returns the declared package a command belongs to.
#
#  * K<::Simple::Package::information requisites> package ?version?
#    Returns a declared package list of requisites.
#
#  * K<::Simple::Package::information required> package ?version?
#    Returns a declared package list of required packages.
#
#  * K<::Simple::Package::information requiredall> package ?version?
#    Returns a declared package full list of required packages.
#
#  * K<::Simple::Package::information requiredby> package ?version?
#    Returns the list of declared packages which require a package.
#
#  * K<::Simple::Package::information install> package ?version?
#    Returns a declared package install script.
#
#  * K<::Simple::Package::information uninstall> package ?version?
#    Returns a declared package uninstall script.
#
#  * K<::Simple::Package::information aliases> package ?version?
#    Returns a declared package list of aliases.
#
#  * K<::Simple::Package::information declaredexportedcommands> package
#       ?version?
#    Returns a declared package declared list of exported commands.
#
#  * K<::Simple::Package::information exportedcommands> package ?version?
#    Returns a declared package current list of exported commands.
#
#  * K<::Simple::Package::information privatecommands> ?-noclasses? package
#       ?version?
#    Returns a declared package current list of private commands.
#
#  * K<::Simple::Package::information publiccommands> ?-noclasses? package
#       ?version?
#    Returns a declared package current list of public commands.
#
#  * K<::Simple::Package::information commands> ?-noclasses? package ?version?
#    Returns a declared package current list of public and exported commands.
#
#  * K<::Simple::Package::information variables> package ?version?
#    Returns a declared package list of variables.
#
#  * K<::Simple::Package::information options> package ?version?
#    Returns a declared package list of options.
#
#  * K<::Simple::Package::information errors> package ?version?
#    Returns a declared package list of errors.
#
#  * K<::Simple::Package::information types> package ?version?
#    Returns a declared package list of types.
#
#  * K<::Simple::Package::information classes> package ?version?
#    Returns a declared package list of classes.
#
#  * K<::Simple::Package::information exportedcommandsall> package ?version?
#    Returns a declared package current full list of exported commands.
#
#  * K<::Simple::Package::information aliasesall> package ?version?
#    Returns a declared package full list of aliases.
#
#  * K<::Simple::Package::information importedinto> package ?version?
#    Returns the namespaces into which a declared package has been imported.
#
#  * K<::Simple::Package::import> package namespace
#    Imports a package exported commands into a namespace.
#
#  * K<::Simple::Package::unimport> package namespace
#    Unimports a package exported commands from a namespace.
#
#  * K<::Simple::Package::uninstall> package
#    Uninstalls a package.
#
#  * K<::Simple::Package::undeclare> package
#    Undeclares a package.
#
#  * K<::Simple::Package::uninstall-and-undeclare> package
#    Uninstalls and undeclares a package.
#
#  * K<::Simple::Package::forget> package ?version?
#    Forgets a package version.
#
#  * K<::Simple::Package::resync>
#    Resynchronizes the information about available packages.
#
# -history    :
#
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimplePackage-extra 1.0 -firsttimeinstall {

### ===========================================================================
### -command  : ::Simple::Package::information
# -synopsis   : Base command for package K<information> subcommands.
# -access     : private
# -overview   :
#
#     This is the base command for the K<::Simple::Package::information>
#  subcommand procedures.
#
# -arguments  :
#  { args             list           {Subcommand and subcommand arguments}}
#
proc ::Simple::Package::information {
   args
} {
   #========= %BASE-COMMAND%

   ### Get the subcommand and full command names
   set subcommand [lindex $args 0]
   set fullCommand ::Simple::Package::information\ $subcommand

   ### Evaluate the subcommand procedure
   ### Error from subcommand procedure
   #   The K<uplevel> here ensures that the call to the base command
   #   is hidden to the command procedure body
   if {[catch {uplevel 1 [list $fullCommand] [lrange $args 1 end]} result]} {

      ### Compose the usage string
      #   This knows about all subcommands, both those defined in this
      #   package and those defined in P<SimpleDeclare>
      switch -exact -- $subcommand {
         aliases -
         aliasesall -
         api -
         classes -
         declaredexportedcommands -
         errors -
         exportedcommands -
         exportedcommandsall -
         importedinto -
         install -
         metadata -
         namespaces -
         options -
         required -
         requiredall -
         requiredby -
         requisites -
         state -
         types -
         uninstall -
         variables {
            set usage {package ?version?}
         }
         collisions {
            set usage {package ?namespace?}
         }
         commands -
         privatecommands -
         publiccommands {
            set usage {-noclasses package ?version?}
         }
         declared -
         installed -
         known -
         uninstalled {
            set usage ?pattern?
         }
         optionchoices -
         optiondescription -
         optiontype -
         optionvalue {
            set usage {package version flag}
         }
         packagealias {
            set usage alias
         }
         packageclass {
            set usage class
         }
         packagecommand {
            set usage command
         }
         packagenamespace {
            set usage namespace
         }
         default {
            set usage {}
         }
      }

      ### Process the subcommand procedure error
      ::Simple::Error::Priv::subcommand-error\
         ::Simple::Package::information $result $usage\
         [lsort $::Simple::Package::InformationSubcommands]

   ### No error from subcommand procedure
   } else {

      ### Return the subcommand procedure return value
      return $result
   }
}

### ===========================================================================
### -command  : ::Simple::Package::information known
# -synopsis   : Returns the list of known packages.
# -overview   :
#
#     This procedure returns the list of known packages (that is, those for
#  which K<package ifneeded> and/or K<package provide> has been invoked) and
#  whose name matches the given pattern.
#
#     The returned packages can be in any state but A<unknown>, that is,
#  A<available>, A<declared> or A<installed>.
#
# -arguments  :
#  {?pattern?         pattern {}     {Pattern}}
#
# -returns :  
#
#     The list of known packages whose name matches the given pattern in the
#  format ||{package version}||.
#
# -remarks    :
#
#  * Use no pattern to get all packages.
#
proc {::Simple::Package::information known} {
   {pattern {}}
} {
   ### Package pattern given
   if {[string compare $pattern {}]} {

      ### Get packages matching the given pattern
      set packages [list]
      foreach package [package names] {
         if {[string match $pattern $package]} {
            lappend packages $package
         }
      }

   ### Package pattern not given
   } else {

      ### Get all packages
      set packages [package names]
   }

   ### Loop over packages
   set answer [list]
   foreach package $packages {

      ### Get the package known versions
      set knownVersions [::Simple::Package::Priv::known-versions $package]

      ### Build the package list
      foreach knownVersion $knownVersions {
         lappend answer [list $package $knownVersion]
      }
   }

   ### Return the list of known packages
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::known-versions
# -synopsis   : Returns a package list of known versions.
# -arguments  :
#  { package          package        {Package name}}
#
# -returns :  The package list of known versions.
#
proc ::Simple::Package::Priv::known-versions {
   package
} {
   ### There is a provided version
   set providedVersion [package provide $package]
   set ifneededVersions [package versions $package]
   if {[string compare $providedVersion {}]} {

      ### Provided and ifneeded versions
      #   But provided version has precedence over equivalent ifneeded version
      set answer $providedVersion
      foreach ifneededVersion $ifneededVersions {
         if {[package vcompare $ifneededVersion $providedVersion]} {
            lappend answer $ifneededVersion
         }
      }

   ### There is no provided version
   } else {

      ### ifneeded versions only
      set answer $ifneededVersions
   }

   ### Return the package list of known version
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::information declared
# -synopsis   : Returns the list of declared packages.
# -overview   :
#
#     This procedure returns the list of declared packages (that is, those for
#  which K<::Simple::Package::declare> has been invoked) and whose name
#  matches the given pattern.
#
#  The returned packages can be in the A<declared> or A<installed> states.
#
# -arguments  :
#  {?pattern?         pattern {}     {Pattern}}
#
# -returns :  
#
#     The list of declared packages whose name matches the given pattern in the
#  format ||{package version}||.
#
# -remarks    :
#
#  * Use no pattern to get all declared packages.
#
proc {::Simple::Package::information declared} {
   {pattern {}}
} {
   ### Package pattern given
   if {[string compare $pattern {}]} {

      ### Get declared packages matching the given pattern
      set packages [list]
      foreach packagePair [array names ::Simple::Package::DeclaredPackages] {
         foreach {package version} $packagePair break
         if {[string match $pattern $package]} {
            lappend packages $packagePair
         }
      }
      return $packages

   ### Package pattern not given
   } else  {

      ### Get all declared packages
      array names ::Simple::Package::DeclaredPackages
   }
}

### ===========================================================================
### -command  : ::Simple::Package::information uninstalled
# -synopsis   : Returns the list of uninstalled packages.
# -overview   :
#
#     This procedure returns the list of uninstalled packages (that is, those
#  for which K<::Simple::Package::declare> has been invoked but which have not
#  been installed yet via K<::Simple::Package::install>) and whose name
#  matches the given pattern.
#
#     All returned packages are in the A<declared> state.
#
# -arguments  :
#  {?pattern?         pattern {}     {Pattern}}
#
# -returns :  
#
#     The list of uninstalled packages whose name matches the given pattern in
#  the format ||{package version}||.
#
# -remarks    :
#
#  * Use no pattern to get all uninstalled packages.
#
proc {::Simple::Package::information uninstalled} {
   {pattern {}}
} {
   ### Get all declared packages matching the given pattern
   set packages [{::Simple::Package::information declared} $pattern]

   ### Loop over packages
   set answer [list]
   foreach packagePair $packages {

      ### Uninstalled package
      if {![info exists ::Simple::Package::InstalledPackages($packagePair)]} {

         ### Add the package to the list of uninstalled packages
         lappend answer $packagePair
      }
   }

   ### Return the list of uninstalled packages
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::information installed
# -synopsis   : Returns the list of installed packages.
# -overview   :
#
#     This procedure returns the list of installed packages (that is, those
#  whose state is A<installed>, either by having being required and not
#  declared or required, declared and also installed) and whose name matches
#  the given pattern.
#
#     All returned packages are in the A<installed> state.
#
# -arguments  :
#  {?pattern?         pattern {}     {Pattern}}
#
# -returns :  
#
#     The list of installed packages whose name matches the given pattern in
#  the format ||{package version}||.
#
# -remarks    :
#
#  * Use no pattern to get all installed packages.
#
proc {::Simple::Package::information installed} {
   {pattern {}}
} {
   ### Package pattern given
   if {[string compare $pattern {}]} {

      ### Get packages matching the given pattern
      set packages [list]
      foreach package [package names] {
         if {[string match $pattern $package]} {
            lappend packages $package
         }
      }

   ### Package pattern not given
   } else {

      ### Get all packages
      set packages [package names]
   }

   ### Loop over packages
   set answer [list]
   foreach package $packages {

      ### Provided package
      set version [package provide $package]
      if {[string compare $version {}]} {

         ### Undeclared or declared and installed package
         set packagePair [list $package $version]
         if {![info exists ::Simple::Package::DeclaredPackages($packagePair)]\
            || [info exists\
            ::Simple::Package::InstalledPackages($packagePair)]} {

            ### Add the package to the list of installed packages
            lappend answer $packagePair
         }
      }
   }

   ### Return the list of installed packages
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::delete-declared
# -synopsis   : Deletes a package from the list of declared packages.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::delete-declared {
   packagePair
} {
   unset ::Simple::Package::DeclaredPackages($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::delete-installed
# -synopsis   : Deletes a package from the list of installed packages.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::delete-installed {
   packagePair
} {
   unset ::Simple::Package::InstalledPackages($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::delete-namespaces-and-classes-info
# -synopsis   : Deletes a package namespaces and classes information.
#
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::delete-namespaces-and-classes-info {
   packagePair
} {
   unset ::Simple::Package::PackageNamespaces($packagePair)
   unset ::Simple::Package::PackageClasses($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::delete-namespaces-and-classes
# -synopsis   : Deletes a package namespaces and classes.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
# -remarks    :
#
#  * The namespaces and classes may have been already deleted and no error is
#    thrown.
#
# -effects    :
#
#  * Deletes the package namespaces and classes.
#
proc ::Simple::Package::Priv::delete-namespaces-and-classes {
   packagePair
} {
   foreach namespace $::Simple::Package::PackageNamespaces($packagePair) {
      catch {namespace delete $namespace}
   }
   foreach class $::Simple::Package::PackageClasses($packagePair) {
      catch {namespace delete $class}
   }
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::assert-package-installed
# -synopsis   : Throws an error if a package is not installed.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::assert-package-installed {
   packagePair
} {
   if {![info exists ::Simple::Package::InstalledPackages($packagePair)]} {
      ::Simple::Error::throw ::Simple::Package::UNINSTALLED $packagePair
   }
}

### ===========================================================================
### -command  : ::Simple::Package::information namespaces
# -synopsis   : Returns a declared package namespaces.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -returns    : The package list of namespaces.
#
proc {::Simple::Package::information namespaces} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Return the package namespaces
   return $::Simple::Package::PackageNamespaces($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::information packagealias
# -synopsis   : Returns the declared package an alias belongs to.
# -overview   :
#
#     This procedure returns the declared package the given alias belongs to,
#  or the empty string if either no declared package owns the given alias or
#  the alias does not exist.
#
# -returns    : The declared package an alias belongs to.
# -arguments  :
#  { alias            commandname    {Alias name}}
#
# -remarks    :
#
#  * The aliases considered as belonging to a package are those given to the
#    A<-aliases> flag when the package was declared.
#
proc {::Simple::Package::information packagealias} {
   alias
} {
   ### Loop over declared packages
   foreach declaredPackage [array names ::Simple::Package::DeclaredPackages] {

      ### Loop over package alias
      foreach aliasPair $::Simple::Package::PackageAliases($declaredPackage) {

         ### Alias found
         if {![string compare $alias [lindex $aliasPair 0]]} {

            ### Get the aliased command namespace
            set namespace [namespace qualifiers [lindex $aliasPair 1]]

            ### Return the declared package the alias belongs to
            return\
               [{::Simple::Package::information packagenamespace} $namespace]
         }
      }
   }

   ### No declared package owns the alias
   set answer {}
}

### ===========================================================================
### -command  : ::Simple::Package::information packagecommand
# -synopsis   : Returns the declared package a command belongs to.
# -overview   :
#
#     This procedure returns the declared package the given command belongs
#  to, or the empty string if no declared package owns the given command.
#  It works for either package commands used in their own namespace or for
#  exported commands from other packages.
#
# -returns    : The declared package a command belongs to.
# -arguments  :
#  { command          unqualifiedname
#                                    {Command name}}
# -limitations:
#
#  * The command must exist.
#
proc {::Simple::Package::information packagecommand} {
   command
} {
   ### Get the original command namespace
   set namespace\
      [namespace qualifiers [uplevel 1 namespace origin [list $command]]]

   ### Return the declared package the command belongs to
   return [{::Simple::Package::information packagenamespace} $namespace]
}

### ===========================================================================
### -command  : ::Simple::Package::information requisites
# -synopsis   : Returns a declared package list of requisites.
# -returns    : The package list of requisites.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
proc {::Simple::Package::information requisites} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Return the package package requisites
   return $::Simple::Package::PackageRequisites($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::information required
# -synopsis   : Returns a declared package list of required packages.
# -returns    : 
#
#  The package list of required packages in the  format ||{package version}||.
#
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
proc {::Simple::Package::information required} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Return the package required packages
   return $::Simple::Package::PackageRequiredPackages($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::information requiredall
# -synopsis   : Returns a declared package full list of required packages.
# -overview   :
#
#     This procedure returns the given package list of required packages plus
#  all packages it may require.  It only takes into account the required
#  packages of declared packages.
#
# -returns    :
#
#     The package full list of required packages in the format ||{package
#  version}||.
#
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
# -references :
#
#  * The K<::Simple::Package::Priv::requiredall> command.
#
proc {::Simple::Package::information requiredall} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Return the package full list of required packages
   return [::Simple::Package::Priv::requiredall $packagePair]
}

### ===========================================================================
### -command  : ::Simple::Package::information requiredby
# -synopsis   : Returns the list of declared packages which require a package.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -returns    : 
#
#     The list of declared packages which require the package in the format
#  ||{package version}||.
#
# -remarks    :
#
#  * The given package does not need to be declared.
#
proc {::Simple::Package::information requiredby} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version 1]

   ### Loop over declared packages
   set answer [list]
   foreach declaredPackage [array names ::Simple::Package::DeclaredPackages] {

      ### This package required the given package
      if {[lsearch -exact\
         $::Simple::Package::PackageRequiredPackages($declaredPackage)\
         $packagePair] != -1} {

         ### Add this package to the list of packages
         lappend answer $declaredPackage
      }
   }

   ### Return the list of packages which require the package
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::requiredall-recursive
# -synopsis   : Returns a package full list of required packages.
# -overview   :
#
#     This procedure returns the given package list of required packages plus
#  all packages it may require.  It only takes into account the required
#  packages of declared packages.
#
# -returns    : The package full list of required packages.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
# -references :
#
#  * The K<::Simple::Package::information requiredall> command.
#
#  * The K<::Simple::Package::Priv::requiredall> command.
#
# -details    :
#     This command is recursive.
#
# -remarks    :
#
#  * No provision is made for removing duplicated packages, use
#    K<::Simple::Package::Priv::requiredall> instead.
#
proc ::Simple::Package::Priv::requiredall-recursive {
   packagePair
} {
   ### Loop over required packages
   set answer $::Simple::Package::PackageRequiredPackages($packagePair)
   foreach requiredPackagePair $answer {

      ### The package has been declared
      if {[info exists\
         ::Simple::Package::DeclaredPackages($requiredPackagePair)]} {

         ### Add this required package required packages
         set answer [concat $answer\
            [::Simple::Package::Priv::requiredall-recursive\
            $requiredPackagePair]]
      }
   }

   ### Return the package full list of required packages
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::requiredall
# -synopsis   : Returns a package full list of required packages.
# -overview   :
#
#     This procedure returns the given package list of required packages plus
#  all packages they may require.  It only takes into account the required
#  packages of declared packages.  In effect, it removes the duplicates from
#  the list returned by K<::Simple::Package::Priv::requiredall-recursive>.
#
# -returns    : The package full list of required packages.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
# -references :
#
#  * The K<::Simple::Package::information requiredall> command.
#
#  * The K<::Simple::Package::Priv::requiredall-recursive> command.
#
proc ::Simple::Package::Priv::requiredall {
   packagePair
} {
   foreach requiredPackagePair\
      [::Simple::Package::Priv::requiredall-recursive $packagePair] {

      set tmp($requiredPackagePair) {}
   }
   return [array names tmp]
}

### ===========================================================================
### -command  : ::Simple::Package::information install
# -synopsis   : Returns a declared package install script.
# -returns    : The package install script.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
proc {::Simple::Package::information install} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   # Delete the package install header added by K<declare-package>
   set script $::Simple::Package::PackageInstallScript($packagePair)
   ::Simple::Script::Header::delete script %PRIV%-declare-package

   ### Return the package install script
   return $script
}

### ===========================================================================
### -command  : ::Simple::Package::information uninstall
# -synopsis   : Returns a declared package uninstall script.
# -returns    : The package uninstall script.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
proc {::Simple::Package::information uninstall} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   # Delete the package uninstall trailer added by K<declare-package>
   set script $::Simple::Package::PackageUninstallScript($packagePair)
   ::Simple::Script::Trailer::delete script %PRIV%-declare-package

   ### Return the package uninstall script
   return $script
}

### ===========================================================================
### -command  : ::Simple::Package::information aliases
# -synopsis   : Returns a declared package list of aliases.
# -overview   :
#
#     This procedure returns a package list of aliases as given to the
#  A<-aliases> flag when the package was declared.
#
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -returns    : The package list of aliases.
# -references :
#
#  * The K<::Simple::Package::information aliasesall> command.
#
#  * The K<::Simple::Class::information aliases> command.
#
# -remarks    :
#
#  * Both aliases in package namespaces and classes are taken into account.
#
proc {::Simple::Package::information aliases} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Return the package list of aliases
   return $::Simple::Package::PackageAliases($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::information declaredexportedcommands
# -synopsis   : Returns a declared package declared list of exported commands.
# -overview   :
#
#     This procedure returns a package list of exported commands as given
#  when the package was declared.
#
# -returns    : The package list of declared qualified exported commands.
#
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -references :
#
#  * The K<::Simple::Package::information exportedcommands> command.
#
#  * The K<::Simple::Package::information exportedcommandsall> command.
#
#  * The K<::Simple::Class::information exportedcommands> command.
#
# -remarks    :
#
#  * Both commands in package namespaces and classes are taken into account.
#
#  * The returned list does not contain base commands.
#
#  * The returned list may contain commands with a percent sign in its name
#    assumed to be internal to The Simple Development Library.
#
proc {::Simple::Package::information declaredexportedcommands} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Return the package list of declared qualified exported commands
   return $::Simple::Package::PackageExportedCommands($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::information exportedcommands
# -synopsis   : Returns a declared package current list of exported commands.
# -overview   :
#
#     This procedure returns a package current list of exported commands as
#  given by the K<namespace export> command in each of its namespaces and
#  classes, that is, a snapshot of the currently exported commands is taken. 
#  Thus, the returned list of exported commands might differ from the declared
#  list of exported commands given when the package was declared in case any
#  command was exported after the package declaration or if an extra package
#  exports commands from namespaces belonging to the non-extra package.
#
# -returns    : The package current list of unqualified exported commands.
#
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -references :
#
#  * The K<::Simple::Package::information declaredexportedcommands> command.
#
#  * The K<::Simple::Package::information exportedcommandsall> command.
#
#  * The K<::Simple::Package::Priv::exported-commands> command.
#
# -remarks    :
#
#  * Both commands in package namespaces and classes are taken into account.
#
#  * The returned list may contain base commands.
#
#  * The returned list may contain commands with a percent sign in its name
#    are those are assumed to be internal to The Simple Development Library.
#
#  * The returned list may contain commands exported by an extra package
#    from namespaces belonging to the non-extra package.
#
proc {::Simple::Package::information exportedcommands} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Loop over package namespaces and classes
   set answer [list]
   foreach namespaceOrClass\
      [::Simple::Package::Priv::namespaces-and-classes $packagePair] {

      ### Get the exported commands list from this namespace or class
      namespace eval $namespaceOrClass {
         uplevel 1 [list set exportedCommands [namespace export]]
      }

      ### Add the exported commands to the package list of exported commands
      set answer [concat $answer $exportedCommands]
   }

   ### Return the package list of unqualified exported commands
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::information-commands
# -synopsis   : Returns a declared package current list of requested commands.
# -returns    : The package current list of requested commands.
# -overview   :
#
#     This procedure returns a package current list of requested commands. 
#  The "requested" commands can be private or public depending on which of
#  A<requestPrivate> or A<requestPublic> is true; if both are false, both the
#  public and exported commands re returned.
#
# -arguments  :
#  {-noclasses        boolflag       {Whether not to include class commands}}
#  { package          package        {Package name}}
#  { version          version        {Package version}}
#  { requestPublic    boolean        {Whether to return public commands only}}
#  { requestPrivate   boolean        {Whether to return private commands only}}
#
# -references :
#
#  * The K<::Simple::Package::information privatecommands> command.
#
#  * The K<::Simple::Package::information publiccommands> command.
#
#  * The K<::Simple::Package::information commands> command.
#
# -remarks    :
#
#  * Both commands in package namespaces and classes are taken into account,
#    unless the A<-noclasses> flag is given, case in which only namespaces are
#    considered.
#
#  * The returned list does not contain base commands.
#
#  * The returned list does not contain commands with a percent sign in its
#    name are those are assumed to be internal to The Simple Development
#    Library.
#
#  * The returned list may contain commands created by an extra package
#    in namespaces belonging to the non-extra package.
#
proc {::Simple::Package::Priv::information-commands} {
   arg1
   arg2
   arg3
   requestPublic
   requestPrivate
} {
   ### A<-noclasses> not given
   if {[string compare $arg1 -noclasses]} {
      set noClasses 0
      set package $arg1
      set version $arg2

   ### A<-noclasses> given
   } else {
      set noClasses 1
      set package $arg2
      set version $arg3
   }

   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Loop over package namespaces
   set candidateCommands [list]
   foreach namespace $::Simple::Package::PackageNamespaces($packagePair) {

      ### Include private or non-private namespaces
      set isPrivate [string match *::Priv* $namespace]
      if {($requestPrivate && $isPrivate) || (!$requestPrivate && !$isPrivate)} {

         ### Get the namespace commands
         # fixme: Tcl 8.3b2 allows [info procs ${namespace}::*]
         set script {uplevel 1 [list set commands [info procs]]}
         if {$requestPublic} {
            append script \n\
               {uplevel 1 [list set exportedCommands [namespace export]]}
         }
         namespace eval $namespace $script

         ### Loop over namespace commands
         foreach command $commands {

            ### The command has not been imported from other namespace (except
            ###    for private commands) and is not exported (for public
            ###    commands) and is not a base command
            if {($requestPrivate || [::Simple::Package::Priv::not-imported\
               $namespace $command]) && (!$requestPublic ||\
               [lsearch -exact $exportedCommands $command] == -1) &&\
               ([string first $::Simple::BaseCommandTag\
               [info body ${namespace}::$command]] == -1)} {

               ### Add the command to the candidates list
               lappend candidateCommands ${namespace}::$command
            }
         }
      }
   }

   ### Include package class commands
   if {!$noClasses} {

      ### Access to match
      if {$requestPrivate} {
         set accessToMatch private
      } else {
         set accessToMatch public
      }

      ### Loop over package classes
      foreach class $::Simple::Package::PackageClasses($packagePair) {

         ### Loop over class commands or methods
         foreach commandOrMethod\
            [namespace eval $class {[uplevel 1 set class]::info function}] {


            ### Get this class exported command list
            if {$requestPublic} {
               set exportedCommands [list]
               foreach exportedCommand\
                  [namespace eval $class {namespace export}] {

                  lappend exportedCommands ${class}::$exportedCommand
               }
            }

            ### This class command or method
            if {[string match $class* $commandOrMethod]} {

               ### Get the command or method type and access level
               foreach {commandType commandAccess commandBody}\
                  [namespace eval $class {[uplevel 1 set class]::info function\
                  [uplevel 1 set commandOrMethod] -type -protection -body}] break

               ### Requested command
               #   and not a base command
               if {![string compare $commandType proc] &&\
                  ![string compare $commandAccess $accessToMatch] &&\
                  (!$requestPublic ||\
                  [lsearch -exact $exportedCommands $commandOrMethod] == -1)\
                  && ([string first %$::Simple::BaseCommandTag\
                  $commandBody] == -1)} {

                  ### Add the command to the candidates list
                  lappend candidateCommands $commandOrMethod
               }
            }
         }
      }
   }

   ### Loop over candidate commands
   set answer [list]
   foreach command $candidateCommands {

      ### Not internal command
      if {[::Simple::Priv::is-not-internal $command]} {

         ### Add the command to the package list of requested commands
         lappend answer $command
      }
   }

   ### Return the package list of qualified requested commands
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::information privatecommands
# -synopsis   : Returns a declared package current list of private commands.
# -returns    : The package current list of qualified private commands.
# -overview   :
#
#     This procedure returns a package current list of private commands.  A
#  private command is either one living in a private namespace (that is, one
#  whose name matches the pattern ||*::Priv*||), or a private class command.
#
# -arguments  :
#  {-noclasses        boolflag       {Whether not to include class commands}}
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
# -references :
#
#  * The K<::Simple::Package::Priv::information-commands> command.
#
# -remarks    :
#
#  * Both commands in package namespaces and classes are taken into account,
#    unless the A<-noclasses> flag is given, case in which only namespaces are
#    considered.
#
#  * The returned list does not contain base commands.
#
#  * The returned list does not contain commands with a percent sign in its
#    name are those are assumed to be internal to The Simple Development
#    Library.
#
#  * The returned list may contain commands created by an extra package
#    in namespaces belonging to the non-extra package.
#
proc {::Simple::Package::information privatecommands} {
   arg1
   {arg2 %DEFAULT%}
   {arg3 %DEFAULT%}
} {
   ### Return the package list of qualified private commands
   return [::Simple::Package::Priv::information-commands $arg1 $arg2 $arg3 0 1]
}

### ===========================================================================
### -command  : ::Simple::Package::information publiccommands
# -synopsis   : Returns a declared package current list of public commands.
# -returns    : The package current list of qualified public commands.
#
# -arguments  :
#  {-noclasses        boolflag       {Whether not to include class commands}}
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
# -references :
#
#  * The K<::Simple::Package::Priv::information-commands> command.
#
# -remarks    :
#
#  * Both commands in package namespaces and classes are taken into account,
#    unless the A<-noclasses> flag is given, case in which only namespaces are
#    considered.
#
#  * The returned list does not contain base commands.
#
#  * The returned list does not contain commands with a percent sign in its
#    name are those are assumed to be internal to The Simple Development
#    Library.
#
#  * The returned list may contain commands created by an extra package
#    in namespaces belonging to the non-extra package.
#
proc {::Simple::Package::information publiccommands} {
   arg1
   {arg2 %DEFAULT%}
   {arg3 %DEFAULT%}
} {
   ### Return the package list of qualified public commands
   return [::Simple::Package::Priv::information-commands $arg1 $arg2 $arg3 1 0]
}

### ===========================================================================
### -command  : ::Simple::Package::information commands
# -synopsis   : Returns a declared package current list of public and exported commands.
# -returns    : 
#
#  The package current list of qualified public and exported commands.
#
# -overview   :
#
#     This procedure returns the given package list of commands currently
#  available to the user.  This list includes both the exported and public
#  commands, both in namespaces and classes.  
#
# -arguments  :
#  {-noclasses        boolflag       {Whether not to include class commands}}
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -references :
#
#  * The K<::Simple::Package::information api> command.
#
#  * The K<::Simple::Class::information commands> command.
#  
#  * The K<::Simple::Package::Priv::information-commands> command.
#
# -remarks    :
#
#  * Both commands in package namespaces and classes are taken into account,
#    unless the A<-noclasses> flag is given, case in which only namespaces are
#    considered.
#
#  * The returned list does not contain base commands.
#
#  * The returned list does not contain commands with a percent sign in its
#    name are those are assumed to be internal to The Simple Development
#    Library.
#
#  * The returned list may contain commands created by an extra package
#    in namespaces belonging to the non-extra package.
#
proc {::Simple::Package::information commands} {
   arg1
   {arg2 %DEFAULT%}
   {arg3 %DEFAULT%}
} {
   ### Return the package list of qualified public and exported commands
   return [::Simple::Package::Priv::information-commands $arg1 $arg2 $arg3 0 0]
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::not-imported
# -synopsis   : Returns whether a command has not been imported from other namespace
# -returns    : 
#
#  Whether the given command has not been imported from other namespace
#
# -overview   :
#
#     This procedure returns whether the given command has not been imported
#  from other namespace.  It is needed because Tcl 8.2.1 and beyond includes
#  imported commands in the result of [info procs]
#
# -arguments  :
#  { namespace        namespace      {Namespace}}
#  { command          unqualifiedname
#                                    {Command name}}
#
proc ::Simple::Package::Priv::not-imported {
   namespace
   command
} {
   return [expr {![string compare $namespace [namespace qualifiers\
      [namespace origin ${namespace}::$command]]]}]
}

### ===========================================================================
### -command  : ::Simple::Package::information variables
# -synopsis   : Returns a declared package list of variables.
# -returns    : The package list of variables.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -remarks    :
#
#  * The variables considered as belonging to a package are those given to
#    the A<-variable> flag when the package was declared with
#    K<declare-package>.
#
# -references :
#
#  * The K<::Simple::Package::information api> command.
#
proc {::Simple::Package::information variables} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Return the package list of variables
   return $::Simple::Package::PackageVariables($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::information options
# -synopsis   : Returns a declared package list of options.
# -returns    : The package list of options.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -remarks    :
#
#  * The options considered as belonging to a package are those given to the
#    A<-option> flag when the package was declared with K<declare-package>.
#
# -references :
#
#  * The K<::Simple::Package::information api> command.
#
proc {::Simple::Package::information options} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Return the package list of options
   return $::Simple::Package::PackageOptions($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::information errors
# -synopsis   : Returns a declared package list of errors.
# -returns    : The package list of errors.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -remarks    :
#
#  * The errors considered as belonging to a package are those given to the
#    A<-error> flag when the package was declared with K<declare-package>.
#
# -references :
#
#  * The K<::Simple::Package::information api> command.
#
proc {::Simple::Package::information errors} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Return the package list of errors
   return $::Simple::Package::PackageErrors($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::information types
# -synopsis   : Returns a declared package list of types.
# -returns    : The package list of types.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -remarks    :
#
#  * The types considered as belonging to a package are those given to the
#    A<-type> flag when the package was declared with K<declare-package>.
#
# -references :
#
#  * The K<::Simple::Package::information api> command.
#
proc {::Simple::Package::information types} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Return the package list of types
   return $::Simple::Package::PackageTypes($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::information classes
# -synopsis   : Returns a declared package list of classes.
# -returns    : The package list of classes.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -remarks    :
#
#  * The classes considered as belonging to a package are those given to the
#    A<-class> flag when the package was declared with K<declare-package>.
#
# -references :
#
#  * The K<::Simple::Package::information api> command.
#
proc {::Simple::Package::information classes} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Return the package list of classes
   return $::Simple::Package::PackageClasses($packagePair)
}

### ===========================================================================
### -command  : ::Simple::Package::information exportedcommandsall
# -synopsis   : Returns a declared package current full list of exported commands.
# -overview   :
#
#     This procedure returns a package current list of exported commands plus
#  those exported from all packages it may require.  It only takes into
#  account the required packages of declared packages.
#
#     For each package, the list of exported commands is as given by the
#  K<namespace export> command in each of its namespaces, that is, a snapshot
#  of the exported commands is taken.  Thus, the returned list of exported
#  commands might differ from the declared list of exported commands given
#  when the package was declared in case any command was exported after the
#  package declaration or if an extra package exports commands from namespaces
#  belonging to the non-extra package.
#
# -returns    : The package current full list of unqualified exported commands.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -references :
#
#  * The K<::Simple::Package::information declaredexportedcommands> command.
#
#  * The K<::Simple::Package::information exportedcommands> command.
#
# -remarks    :
#
#  * Both commands in package namespaces and classes are taken into account.
#
#  * The returned list may contain base commands.
#
#  * The returned list may contain commands with a percent sign in its name
#    are those are assumed to be internal to The Simple Development Library.
#
#  * The returned list may contain commands exported by extra packages
#    from namespaces belonging to non-extra packages.
#
proc {::Simple::Package::information exportedcommandsall} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Get all required packages
   set requiredPackagePairs\
      [::Simple::Package::Priv::requiredall $packagePair]
   lappend requiredPackagePairs $packagePair

   ### Loop over required packages
   set answer [list]
   foreach requiredPackagePair $requiredPackagePairs {

      ### The package has been declared
      if {[info exists\
         ::Simple::Package::DeclaredPackages($requiredPackagePair)]} {

         ### Loop over qualified package exported commands
         foreach command\
            [::Simple::Package::Priv::exported-commands $requiredPackagePair] {

            ### Add the command to the package list of exported commands
            lappend answer [namespace tail $command]
         }
      }
   }

   ### Return the package full list of unqualified exported commands
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::information aliasesall
# -synopsis   : Returns a declared package full list of aliases.
# -overview   :
#
#     This procedure returns the given package list of aliases plus those from
#  all packages it may require.  It only takes into account the required
#  packages of declared packages.
#
#     For each package, the list of aliases is as given to the A<-aliases>
#  flag when the package was declared.
#
# -returns    : The package full list of aliases.
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -references :
#
#  * The K<::Simple::Package::information aliases> command.
#
# -remarks    :
#
#  * Both aliases in package namespaces and classes are taken into account.
#
proc {::Simple::Package::information aliasesall} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Get all required packages
   set requiredPackagePairs [::Simple::Package::Priv::requiredall $packagePair]
   lappend requiredPackagePairs $packagePair

   ### Loop over required packages
   set answer [list]
   foreach requiredPackagePair $requiredPackagePairs {

      ### The package has been declared
      if {[info exists\
         ::Simple::Package::DeclaredPackages($requiredPackagePair)]} {

         ### Add this package aliases to the full list of aliases
         set answer [concat $answer\
            $::Simple::Package::PackageAliases($requiredPackagePair)]
      }
   }

   ### Return the package full list aliases
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::information importedinto
# -synopsis   : Returns the namespaces into which a declared package has been imported.
# -overview   :
#
#     This procedure returns the list of namespaces into which the given
#  package has been imported or the empty list for uninstalled packages.
#
# -returns    : 
#
#     The list of namespaces into which the package has been imported or the
#  empty list for uninstalled packages.
#
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
proc {::Simple::Package::information importedinto} {
   package
   {version %DEFAULT%}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### The package is installed
   set answer [list]
   if {[info exists ::Simple::Package::InstalledPackages($packagePair)]} {

      ### Add the namespaces into which the package has been imported
      foreach element [array names ::Simple::Package::PackageImportedInto\
         $packagePair%*] {

         lappend answer $::Simple::Package::PackageImportedInto($element)
      }
   }

   ### Return the list of namespaces into which the package has been imported
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::delete-importedinto
# -synopsis   : Deletes a package list of namespaces imported into.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#  {?namespace?       namespace %DEFAULT%   
#                                    {Namespace}}
#
proc ::Simple::Package::Priv::delete-importedinto {
   packagePair
   {namespace %DEFAULT%}
} {
   ### No namespace given
   if {![string compare $namespace %DEFAULT%]} {

      ### Delete the package list of namespaces imported into
      foreach element [array names ::Simple::Package::PackageImportedInto\
         $packagePair%*] {

         unset ::Simple::Package::PackageImportedInto($element)
      }
   } else {

      ### Delete the namespace from the list of namespaces imported into
      unset ::Simple::Package::PackageImportedInto($packagePair%$namespace)
   }
}

### ===========================================================================
### -command  : ::Simple::Package::import
# -synopsis   : Imports a package exported commands into a namespace.
# -overview   :
#
#     This procedure imports the given installed package exported commands
#  into the given namespace.
#
#     An error is thrown if the package is not installed.
#
#     An error is thrown if the package has already been imported into the
#  given namespace.  This implies that the namespace can not be the global
#  one, as any packages is imported into the global namespace upon
#  installation.
#
#     An error is thrown if importing the package exported commands into the
#  given namespace would result in a collision; use
#  K<::Simple::Package::information collisions> to ensure there are no
#  collisions.
#
# -arguments  :
#  { package          package        {Package name}}
#  { namespace        namespace      {Namespace}}
#
# -returns    : The empty string.
# -effects    :
#
#  * Imports a package exported commands into the given namespace.
#
proc ::Simple::Package::import {
   package
   namespace
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package %DEFAULT%]

   ### Assert the package is installed
   ::Simple::Package::Priv::assert-package-installed $packagePair

   ### Assert the namespace exists
   ::Simple::Package::Priv::assert-namespace-exists $namespace

   ### The package has not been imported into the namespace
   if {![info exists\
      ::Simple::Package::PackageImportedInto($packagePair%$namespace)]} {

      ### Assert no collision will occur when importing the package
      ::Simple::Package::Priv::assert-no-collisions $packagePair\
         $namespace

      ### Import the package into the namespace
      ::Simple::Package::Priv::do-import $packagePair $namespace

   ### The package has already been imported into the namespace
   } else {
      ::Simple::Error::throw ::Simple::Package::ALREADY-IMPORTED\
         $packagePair $namespace
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Package::unimport
# -synopsis   : Unimports a package exported commands from a namespace.
# -overview   :
#
#     This procedure unimports the given package exported commands from the
#  given namespace.
#
#     An error is thrown if the package is not installed.
#
#     An error is thrown if the package has not been imported into the
#  given namespace.
#
# -arguments  :
#  { package          package        {Package name}}
#  { namespace        namespace      {Namespace}}
#
# -returns    : The empty string.
# -effects    :
#
#  * Unimports the package exported commands from the given namespace.
#
proc ::Simple::Package::unimport {
   package
   namespace
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package %DEFAULT%]

   ### Assert the package is installed
   ::Simple::Package::Priv::assert-package-installed $packagePair

   ### Assert the namespace exists
   ::Simple::Package::Priv::assert-namespace-exists $namespace

   ### The package has been imported into the namespace
   if {[info exists\
      ::Simple::Package::PackageImportedInto($packagePair%$namespace)]} {

      ### Unimport the package from the namespace
      ::Simple::Package::Priv::do-unimport $packagePair $namespace

   ### The package has not been imported into the namespace
   } else {

      ### Throw error
      ::Simple::Error::throw ::Simple::Package::NOT-IMPORTED\
         $packagePair $namespace
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::do-unimport
# -synopsis   : Unimports a package exported commands from a namespace.
# -overview   :
#
#     This procedure loops over a package namespaces and classes and unimports
#  their exported commands from the given namespace.
#
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#  { namespace        namespace      {Namespace}}
#
# -effects    :
#
#  * Unimports a package exported commands from the given namespace.
#
proc ::Simple::Package::Priv::do-unimport {
   packagePair
   namespace
} {
   ### Unimport the package exported commands from the namespace
   set namespacesAndClasses\
      [::Simple::Package::Priv::namespaces-and-classes $packagePair]
   namespace eval $namespace {
      foreach packageNamespace [uplevel 1 [list set namespacesAndClasses]] {
         namespace forget ${packageNamespace}::*
      }
   }

   ### Remove the namespace from the package list of namespaces imported into
   ::Simple::Package::Priv::delete-importedinto $packagePair $namespace
}

### ===========================================================================
### -command  : ::Simple::Package::uninstall
# -synopsis   : Uninstalls a package.
# -overview   :
#
#     This procedure uninstalls a package previously installed via the
#  K<::Simple::Package::install> command.  The following tasks are performed
#  in this same order:
#
#  * All package exported commands are unimported from all namespaces they
#    were imported into.
#
#  * The package uninstall script is executed in the global namespace.
#
#  * All package aliases are deleted.
#
#  An error is thrown if the package is not installed.
#
# -arguments  :
#  { package          package        {Package name}}
#
# -returns    : The empty string.
# -remarks    :
#
#  * Notice that unlike K<::Simple::Package::install> which installs all
#    packages required by the package being installed,
#    K<::Simple::Package::uninstall> uninstalls the given package only.
#
# -effects    :
#
#  * Those of the evaluated script.
#
#  * Deletes the package aliases.
#
#  * Deletes the objects of the classes of object packages declared with
#    K<declare-package>.
#
#  * Unimports the package exported commands from the namespaces they were
#    imported into.
#
proc ::Simple::Package::uninstall {
   package
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package %DEFAULT%]

   ### Assert the package is installed
   ::Simple::Package::Priv::assert-package-installed $packagePair

   ### Error evaluating the package uninstall script
   if {[catch {namespace eval ::\
      $::Simple::Package::PackageUninstallScript($packagePair)} result]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Package::ERROR-EVAL-SCRIPT\
         $packagePair uninstall $result

   ### No error evaluating the package uninstall script
   } else {

      ### Delete the objects of the classes of object packages
      #   The K<catch> guards against P<SimpleClass> not being installed
      catch {::Simple::Class::Priv::delete-package-objects $packagePair}

      ### Unimport the package from the namespaces it has been imported into
      foreach element\
         [array names ::Simple::Package::PackageImportedInto $packagePair%*] {

         ::Simple::Package::Priv::do-unimport\
            $packagePair $::Simple::Package::PackageImportedInto($element)
      }

      ### Hide the package types
      if {[llength $::Simple::Package::PackageTypes($packagePair)] > 0} {
         ::Simple::Type::Priv::hide\
            $::Simple::Package::PackageTypes($packagePair)
      }

      ### Delete the package aliases
      foreach aliasPair $::Simple::Package::PackageAliases($packagePair) {
         interp alias {} [lindex $aliasPair 0] {}
      }

      ### Delete this package from the list of installed packages
      ::Simple::Package::Priv::delete-installed $packagePair
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Package::undeclare
# -synopsis   : Undeclares a package.
# -overview   :
#
#  This procedure undeclares a declared package.
#
#     An error is thrown if the package is installed; use
#  K<::Simple::Package::uninstall> to uninstall a package.
#
# -arguments  :
#
#  { package          package        {Package name}}
#
# -returns    : The empty string.
# -remarks    :
#
#  * Notice that unlike K<::Simple::Package::declare> which requires all
#    packages required by the package being declared,
#    K<::Simple::Package::undeclare> undeclares the given package only.
#
# -effects    :
#
#  * Deletes the package namespaces.
#
#  * Deletes the classes of object packages declared with K<declare-package>.
#
#  * Deletes the types of packages declared with K<declare-package>.
#
#  * Forgets a package, but keeps the K<ifneeded> script, if any.
#
proc ::Simple::Package::undeclare {
   package
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package %DEFAULT%]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Assert the package is not installed
   ::Simple::Package::Priv::assert-package-not-installed $packagePair

   ### Delete the package namespaces and classes
   ::Simple::Package::Priv::delete-namespaces-and-classes $packagePair

   ### Delete the package namespace and classes information
   ::Simple::Package::Priv::delete-namespaces-and-classes-info $packagePair

   ### Delete the package variables information
   unset ::Simple::Package::PackageVariables($packagePair)

   ### Delete the package options information
   unset ::Simple::Package::PackageOptions($packagePair)

   ### Delete the package errors information
   unset ::Simple::Package::PackageErrors($packagePair)

   ### Delete the package types
   foreach type $::Simple::Package::PackageTypes($packagePair) {
      #fixme: the [catch] is needed because all package types are tried
      #       to be deleted, including the class types, but the classes
      #       along their class types have been already deleted above
      catch {::Simple::Type::Priv::delete-with-hidden $type 1}
   }

   ### Delete the package types information
   unset ::Simple::Package::PackageTypes($packagePair)

   ### Delete the package requisites information
   unset ::Simple::Package::PackageRequisites($packagePair)

   ### Delete the package required packages information
   unset ::Simple::Package::PackageRequiredPackages($packagePair)

   ### Delete the package list of namespaces imported into
   ::Simple::Package::Priv::delete-importedinto $packagePair

   ### Delete the package exported commands information
   unset ::Simple::Package::PackageExportedCommands($packagePair)

   ### Delete the package aliases information
   unset ::Simple::Package::PackageAliases($packagePair)

   ### Delete the package install script
   unset ::Simple::Package::PackageInstallScript($packagePair)

   ### Delete the package first time install script
   unset ::Simple::Package::PackageFirstInstallScript($packagePair)

   ### Delete the package uninstall script
   unset ::Simple::Package::PackageUninstallScript($packagePair)

   ### Delete the package metadata
   unset ::Simple::Package::PackageMetadata($packagePair)

   ### Delete this package from the list of declared packages
   ::Simple::Package::Priv::delete-declared $packagePair

   ### Set the state of the package to available or unknown
   foreach {package version} $packagePair break
   ::Simple::Package::Priv::available-or-unknown $package

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::available-or-unknown
# -synopsis   : Sets the state of a package to available or unknown.
# -overview   :
#
#     This procedure sets the state of a package to available or unknown
#  depending on whether or not there is an ifneeded script for the package.
#
# -arguments  :
#  { package          package        {Package name}}
#
# -effects    :
#
#  * Forgets a package, but keeps its K<ifneeded> script, if any.
#
proc ::Simple::Package::Priv::available-or-unknown {
   package
} {
   ### Loop over package versions
   set versionsNotToForget [package versions $package]
   foreach versionNotToForget $versionsNotToForget {

      ### Save the package ifneeded script
      set ifneededScript($versionNotToForget)\
         [package ifneeded $package $versionNotToForget]
   }

   ### Forget all package versions
   package forget $package

   ### Loop over package versions
   foreach versionNotToForget $versionsNotToForget {

      ### Restore the package ifneeded script
      package ifneeded $package $versionNotToForget\
         $ifneededScript($versionNotToForget)
   }
}

### ===========================================================================
### -command  : ::Simple::Package::uninstall-and-undeclare
# -synopsis   : Uninstalls and undeclares a package.
# -overview   :
#
#     This procedure changes a package state from A<installed> to
#  A<available>.  If the package was declared, it is first uninstalled and
#  then undeclared; otherwise the package is simply unprovided.  Notice that
#  in the later case a full uninstallation might require manually forgeting
#  the package namespaces via K<namespace forget>.
#
#  An error is thrown if the package is not installed.
#
# -arguments  :
#  { package          package        {Package name}}
#
# -returns    : The empty string.
# -remarks    :
#
#  * Notice that unlike K<::Simple::Package::install> which installs all
#    packages required by the package being installed,
#    K<::Simple::Package::uninstall-and-undeclare> uninstalls the given
#    package only.
#
# -effects    :
#
#  * Forgets a package, but keeps its K<ifneeded> script, if any.
#
proc ::Simple::Package::uninstall-and-undeclare {
   package
} {
   ### Undeclared package
   if {[llength [array names\
      ::Simple::Package::DeclaredPackages "$package *"]] == 0} {

      ### Assert the package is installed
      set version [package provide $package]
      if {![string compare $version {}]} {
         ::Simple::Error::throw ::Simple::Package::UNINSTALLED $package
      }

      ### Set the state of the package to available or unknown
      ::Simple::Package::Priv::available-or-unknown $package

   ### Declared package
   } else {

      ### Uninstall the package
      ::Simple::Package::uninstall $package

      ### Undeclare the package
      ::Simple::Package::undeclare $package
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Package::forget
# -synopsis   : Forgets a package version.
# -overview   :
#
#     This command is similar to K<package forget> but accepts an aditional
#  second argument which, when given, forgets that package version only.
#
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -returns    : The empty string.
# -effects    :
#
#  * Forgets all versions of a package or a single version if given.
#
proc ::Simple::Package::forget {
   package
   {version %DEFAULT%}
} {
   ### Version given
   if {[string compare $version %DEFAULT%]} {

      ### Get the version equivalent to that given
      set version\
         [::Simple::Package::Priv::equivalent-version $package $version]

      ### Get the package known versions
      set knownVersions [::Simple::Package::Priv::known-versions $package]

      ### Package version known
      set index [lsearch -exact $knownVersions $version]
      if {$index != -1} {

         ### Loop over package versions not to forget
         set versionsNotToForget [lreplace $knownVersions $index $index]
         foreach versionNotToForget $versionsNotToForget {

            ### Save the package ifneeded script
            set ifneededScript($versionNotToForget)\
               [package ifneeded $package $versionNotToForget]
         }

         ### Save the package provided version
         set providedVersion [package provide $package]

         ### Forget all package versions
         package forget $package

         ### Loop over package versions not to forget
         foreach versionNotToForget $versionsNotToForget {

            ### Restore the package ifneeded script
            package ifneeded $package $versionNotToForget\
               $ifneededScript($versionNotToForget)
         }

         ### Restore the package provided version, if
         ###    any and if it is not the forgotten one
         if {[string compare $providedVersion {}] &&\
            [package vcompare $version $providedVersion]} {

            package provide $package $providedVersion
         }
      }

   ### Version not given
   } else {

      ### Forget all package versions
      package forget $package
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Package::resync
# -synopsis   : Resynchronizes the information about available packages.
# -overview   :
#
#     This command resynchronizes the information about available packages by
#  invoking the K<unknown> procedure.
#
# -returns    : The empty string.
# -history    :
#
#  10-may-2000   Implementation by Don Porter (donald.porter@nist.gov)
#
proc ::Simple::Package::resync {
} {
   ### Resynchronize the information about available packages
   namespace eval :: [package unknown] Tcl [package provide Tcl]

   ### Return the empty string
   return {}
}
}
