### ===========================================================================
### -package  : SimpleProgram-extra
#
package require SimpleDevLibPackage 1.0
declare-package {

   SimpleProgram-extra

} -version {

   1.0

} -project {

   The Simple Development Library.

} -synopsis {

   The Simple Development Library program declaration.

} -module {

   extra

} -api {

   * K<::Simple::Program::information options>
     Returns the program options.

   * K<::Simple::Program::information types>
     Returns the program types.

   * K<::Simple::Program::information errors>
     Returns the program errors.

   * K<::Simple::Program::information examples>
     Returns the program examples.

   * K<::Simple::Program::information synopsis>
     Returns the program synopsis.

   * K<::Simple::Program::information name>
     Returns the program name.

   * K<::Simple::Program::information help>
     Returns the program help message.

   * K<::Simple::Program::information commands>
     Returns the program commands.

   * K<::Simple::Program::information metadata>
     Returns the program metadata.

   * K<::Simple::Program::information version>
     Returns the program version.

   * K<::Simple::Program::information variables>
     Returns the program variables.

} -history {

   22-jun-2005   The Simple Development Library version 1.0

} -copyright {

   Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)

} -command {

### ===========================================================================
### -command  : ::Simple::Program::information metadata

{
   ::Simple::Program::information metadata

} -synopsis {

   Returns the program metadata.

} -overview {

      This procedure returns a list of pairs suitable to be given as argument
   to the K<array set> command.  Each element is one of the flags given to
   K<declare-program> upon program declaration (with no hypen), but for
   A<-body> (which is not stored), the item declarations (A<-error>, A<-type>,
   A<-variable>, A<-command> and A<-option>; these are obtained using the
   corresponding K<::Simple::Program::information> subcommand), the
   A<-optionsusage> flag (use K<::Simple::Option::information usage> to get
   this one) and A<-version>, A<-synopsis>, A<-help>, A<-examples>,
   A<-requisites> and A<-name> for which specific
   K<::Simple::Program::information> subcommands exists.

      Thus, the actual list of flags is:  A<-project>, A<-overview>,
   A<-usage>, A<-exitstatus>, A<-keywords>, A<-assumptions>, A<-resources>,
   A<-limitations>, A<-details>, A<-references>, A<-remarks>, A<-todo>,
   A<-cmid>, A<-sprs>, A<-smrs>, A<-history> and A<-copyright>.
   
      Notice that no data is returned unless The Simple Development Library
   A<-storemetadata> option is set when the program was declared. 
   
      If a particular flag was not given in the program declaration, the empty
   string is returned as that flag contents.

} -returns {

   The program metadata as an item/contents pairs list.

} -body {

   ### The program metadata was stored
   set answer [list]
   if {[llength $::Simple::Program::Metadata] > 0} {
      foreach flag [list project overview usage exitstatus keywords\
         assumptions resources limitations details references remarks todo\
         cmid sprs smrs history copyright]\
         contents $::Simple::Program::Metadata {

         ### Add the metadata to the list
         lappend answer $flag $contents
      }
   }

   ### Return the program metadata
   return $answer
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::information name

{
   ::Simple::Program::information name

} -synopsis {

   Returns the program name.

} -returns {

   The program name.

} -body {

   ### Return the program name
   return $::Simple::Program::Information(name)
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::information version

{
   ::Simple::Program::information version

} -synopsis {

   Returns the program version.

} -returns {

   The program version.

} -body {

   ### Return the program version
   return $::Simple::Program::Information(version)
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::information requisites

{
   ::Simple::Program::information requisites

} -synopsis {

   Returns the program requisites.

} -returns {

   The program requisites.

} -body {

   ### Return the program requisites
   return $::Simple::Program::Information(requisites)
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::information synopsis

{
   ::Simple::Program::information synopsis

} -synopsis {

   Returns the program synopsis.

} -returns {

   The program synopsis.

} -body {

   ### Return the program synopsis
   return $::Simple::Program::Information(synopsis)
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::information help

{
   ::Simple::Program::information help

} -synopsis {

   Returns the program help message.

} -returns {

   The program help message .

} -body {

   ### Ensure the help message is up to date
   ::Simple::Program::Priv::help-message

   ### Return the program help message
   return $::Simple::Program::Information(help)
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::information examples

{
   ::Simple::Program::information examples

} -synopsis {

   Returns the program examples.

} -returns {

   The program examples.

} -body {

   ### Return the program examples
   return $::Simple::Program::Information(examples)
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::information options

{
   ::Simple::Program::information options

} -synopsis {

   Returns the program options.

} -returns {

   The program options.

} -body {

   ### Return the program options
   return $::Simple::Option::PrimaryOptions
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::information errors

{
   ::Simple::Program::information errors

} -synopsis {

   Returns the program errors.

} -returns {

   The program errors.

} -body {

   ### Return the program errors
   return [array names ::%Errors%]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::information variables

{
   ::Simple::Program::information variables

} -synopsis {

   Returns the program variables.

} -returns {

   The program variables.

} -body {

   ### Return the program variables
   return [array names ::%Variables%]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::information types

{
   ::Simple::Program::information types

} -synopsis {

   Returns the program types.

} -returns {

   The program types.

} -body {

   ### Return the program types
   return $::Simple::Program::Types
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::information commands

{
   ::Simple::Program::information commands

} -synopsis {

   Returns the program commands.

} -returns {

   The program commands.

} -body {

   ### Return the program commands
   return $::Simple::Program::Commands
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::Priv::help-message

{
   ::Simple::Program::Priv::help-message

} -synopsis {

   Updates the help message.

} -overview {

      This procedure ensures the program help message is up to date by
   providing a default help message if A<--help> was not provided to
   K<declare-program> or performing

} -returns {

   The empty string.

} -body {

   ### A<--help> not given to K<declare-program>
   if {![string compare $::Simple::Program::Information(help) %DEFAULT%]} {

      ### Provide the default help message
      set help %USAGE%\n\n%SYNOPSIS%\n\n%OPTIONS%

   ### A<--help> given to K<declare-program>
   } else {

      ### Get the help message
      set help $::Simple::Program::Information(help)
   }

   # Compose the command line help message from the options information
   # fixme: the hard-coded formatting below should be handled by
   #        P<SimpleDoc> when available

   # Compose the program usage
   set program $::Simple::Program::Information(name)
   set usage "Usage: $program [::Simple::Option::information usage]"

   # Add the A<--help> and A<--version> options
   append usage "\n       $program --help\n       $program --version"

   # Add the A<--examples> option if there are examples available
   if {[string compare $::Simple::Program::Information(examples) {}]} {
      set examples 1
      append usage "\n       $program --examples"
   } else {
      set examples 0
   }

   # Substitute in the help message the
   # string ''%USAGE%'' by the program usage
   regsub -- %USAGE% $help $usage help

   # Substitute in the help message the
   # string ''%SYNOPSIS%'' by the program synopsis
   regsub -- %SYNOPSIS% $help $::Simple::Program::Information(synopsis) help

   # Get the command line options references and description pairs
   set optionPairs [::Simple::Option::information help]
   lappend optionPairs\
      {{    --help} {display this help and exit}}\
      {{    --version} {output version information and exit}}
   if {$examples} {
      lappend optionPairs\
         {{    --examples} {display some usage examples and exit}}\
   }

   # Compute the longest reference length
   set maxReferenceLength 0
   foreach optionPair $optionPairs {
      set referenceLength [string length [lindex $optionPair 0]]
      if {$referenceLength > $maxReferenceLength} {
         set maxReferenceLength $referenceLength
      }
   }

   # Add the formatted options references and descriptions
   set options {}
   foreach optionPair $optionPairs {
      foreach {reference description} $optionPair break
      append options [format {   %-*s  %s}\
         $maxReferenceLength $reference $description]\n
   }

   # Substitute in the help message the string
   # ''%OPTIONS%'' by the detailed options explanation
   regsub -- %OPTIONS% $help [string trimright $options \n] help

   ### Store the help message
   set ::Simple::Program::Information(help) $help

   ### Return the empty string
   return {}
}
}
